import asyncio
import json
import socket
import urllib.error
import urllib.request
from logging import getLogger
from typing import Optional

from defence360agent.contracts.config import Core
from defence360agent.contracts.messages import Message
from defence360agent.internals.iaid import IndependentAgentIDAPI
from defence360agent.utils.json import ServerJSONEncoder
from defence360agent.api.server import API, APIError

logger = getLogger(__name__)


class SendMessageAPI(API):
    URL = "/api/v2/send-message/{method}"

    def __init__(self, rpm_ver: str, base_url: str = None, executor=None):
        self._executor = executor
        self.rpm_ver = rpm_ver
        self.product_name = ''
        self.server_id = None  # type: Optional[str]
        self.license = {}  # type: dict
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = self._BASE_URL

    def set_product_name(self, product_name: str) -> None:
        self.product_name = product_name

    def set_server_id(self, server_id: Optional[str]) -> None:
        self.server_id = server_id

    def set_license(self, license: dict) -> None:
        self.license = license

    async def send_message(self, message: Message) -> None:
        data2send = {
            'payload': message.payload,
            'rpm_ver': self.rpm_ver,
            'message_id': message.message_id,
            'server_id': self.server_id,
            'name': self.product_name,
        }
        post_data = json.dumps(data2send, cls=ServerJSONEncoder).encode()
        if IndependentAgentIDAPI.is_token_expired():
            await IndependentAgentIDAPI.login()

        headers = {
            "Content-Type": "application/json",
            "X-Auth": IndependentAgentIDAPI.get_token()
        }
        request = urllib.request.Request(
            self.base_url + self.URL.format(
                method=message.method
            ),
            data=post_data,
            headers=headers,
            method='POST'
        )
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(self._executor, _do_request, request)


def _do_request(request: urllib.request.Request) -> None:
    try:
        with urllib.request.urlopen(
            request,
            timeout=Core.DEFAULT_SOCKET_TIMEOUT
        ) as response:
            if response.status != 200:
                raise APIError('status code is {}'.format(response.status))
            result = json.loads(response.read().decode())
            if 'status' not in result:
                raise APIError(
                    'unexpected server response: {!r}'.format(result))
            if result['status'] != 'ok':
                raise APIError(
                    'server error: {}'.format(result.get('msg')))
    except (
        UnicodeDecodeError,
        json.JSONDecodeError,
        urllib.error.URLError,
        socket.timeout,
    ) as e:
        raise APIError(f"request failed {e}") from e
