"""
All the config settings for defence360 in one place
"""
import functools
import logging
import os
import re
from abc import abstractmethod
from bisect import bisect_left, bisect_right
from copy import deepcopy
from datetime import timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    _ProtocolMeta,
)

from cerberus import Validator

from defence360agent.api.integration_conf import IntegrationConfig
from defence360agent.contracts.config_provider import (
    CachedConfigReader,
    ConfigError,
    ConfigReader,
    UserConfigReader,
)
from defence360agent.feature_management.checkers import (
    config_cleanup as fm_config_cleanup,
)
from defence360agent.utils import Singleton, dict_deep_update

logger = logging.getLogger(__name__)

ANTIVIRUS_MODE = False

AGENT_CONF = '../.'

NOTIFY, CLEANUP = 'notify', 'cleanup'

NONE, DAY, WEEK, MONTH = 'none', 'day', 'week', 'month'

DEFAULT_INTENSITY_CPU = 2
DEFAULT_INTENSITY_IO = 2
DEFAULT_INTENSITY_RAM = 2048

DEFAULT_RESOURCE_MANAGEMENT_CPU_LIMIT = 2
DEFAULT_RESOURCE_MANAGEMENT_IO_LIMIT = 2
DEFAULT_RESOURCE_MANAGEMENT_RAM_LIMIT = 500

MODSEC_RULESET_FULL = 'FULL'
MODSEC_RULESET_MINIMAL = 'MINIMAL'
_DOS_DETECTOR_DEFAULT_LIMIT = 250
_DOS_DETECTOR_MIN_LIMIT = 1
_DOS_DETECTOR_MIN_INTERVAL = 1

PORT_BLOCKING_MODE_DENY = 'DENY'
PORT_BLOCKING_MODE_ALLOW = 'ALLOW'

DO_NOT_MODIFY_DISCLAMER = """\
############################################################################
# DO NOT MODIFY THIS FILE!!!                                               #
# USE /etc/sysconfig/imunify360/imunify360.config.d/ TO OVERRIDE DEFAULTS  #
############################################################################
"""
DEFAULT_CONFIG_DISCLAMER = """\
############################################################################
# DO NOT MODIFY THIS FILE!!!                                               #
# USE /etc/sysconfig/imunify360/imunify360.config.d/ TO OVERRIDE DEFAULTS  #
# This is an example of default values only                                #
# Changing this file will have no effect                                   #
############################################################################
"""


def int_from_envvar(var: str, default: int, env: Mapping = os.environ) -> int:
    try:
        return int(env[var])
    except KeyError:
        return default
    except ValueError as e:
        raise ValueError("{}: integer required".format(var)) from e


def bool_from_envvar(
    var: str, default: bool, env: Mapping = os.environ
) -> bool:
    TRUE_VALS = ("true", "t", "yes", "y", "1")
    FALSE_VALS = ("false", "f", "no", "n", "0")
    try:
        val = env[var]
    except KeyError:
        return default
    else:
        val = val.lower()
        if val in TRUE_VALS:
            return True
        if val in FALSE_VALS:
            return False
        raise ValueError(
            "{}: should be one of {}".format(var, TRUE_VALS + FALSE_VALS)
        )


def _self_rel2abs(relpath):
    return os.path.join(os.path.dirname(__file__), relpath)


def conf_rel2abs(relpath):
    return os.path.join(os.path.dirname(_self_rel2abs(AGENT_CONF)), relpath)


def _slurp_file(path):
    """Returns content for existing file, otherwise None"""
    try:
        with open(path, 'r') as f:
            return f.read().strip()
    except OSError:
        return None


def choose_value_from_config(section, option, username=None):
    """
    Choose action for config option by checking EndUser's Imunfiy360
    config and Admin config. Admins config applies only if EndUser
    didn't set the default action
    """
    user_config = ConfigFile(username=username).config_to_dict()
    user_section = user_config.get(section)
    if user_section is not None:
        user_value = user_section.get(option)
        if user_value is not None:
            return user_value, username
    logger.debug("Cannot read %s:%s from user config", section, option)
    root_config = ConfigFile().config_to_dict()
    return root_config[section][option], UserType.ROOT


class FromConfig:
    def __init__(self, section, option=None, config_cls=None):
        self.section = section
        self.option = option
        self._config_cls = config_cls
        self._config_instance = None

    def __get__(self, instance, owner):
        if self._config_instance is None:
            if self._config_cls is None:
                self._config_instance = ConfigFile()
            else:
                self._config_instance = self._config_cls()

        section_value = self._config_instance.config_to_dict()[self.section]
        if self.option is not None:
            return section_value[self.option]
        return section_value


class FromFlagFile:
    _flags_location = "/var/imunify360"

    def __init__(self, flag_name):
        self.flag_file = os.path.join(self._flags_location, flag_name)

    def __get__(self, instance, owner):
        return os.path.isfile(self.flag_file)


def not_less_than(minimum):
    return functools.partial(max, minimum)


def nullable(coerce):
    def f(value):
        if value is None:
            return value
        return coerce(value)

    return f


def list_of_ports(default):
    return {
        'type': 'list',
        'schema': {
            'type': 'string',
            'is_port_or_range': True,
        },
        'default_setter': default,
        'coerce': 'sort_ports',
    }


def proactive_defense_rules(proactive_defense):
    if "php_immunity" in proactive_defense and \
            proactive_defense['php_immunity']:
        proactive_defense['mode'] = 'KILL'
        proactive_defense['blamer'] = True
    return proactive_defense


# FIXME: make this function a constant again
def config_schema_root():
    return {
        'AUTO_WHITELIST': {
            'type': 'dict',
            'schema': {
                'timeout': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'max': 70000,
                    'default': 1440,
                },
                'after_unblock_timeout': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'max': 70000,
                    'default': 1440,
                }
            },
            'default': {},
        },
        'BACKUP_RESTORE': {
            'type': 'dict',
            'schema': {
                'max_days_in_backup': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 90,
                },
                'cl_backup_allowed': {
                    'type': 'boolean',
                    'default': True,
                },
                'cl_on_premise_backup_allowed': {
                    'type': 'boolean',
                    'default': False,
                },
            },
            'default': {},
        },
        'NETWORK_INTERFACE': {
            'type': 'dict',
            'schema': {
                'eth_device': {
                    'type': 'string',
                    'default': None,
                    'nullable': True
                },
                'eth6_device': {
                    'type': 'string',
                    'default': None,
                    'nullable': True
                },
                'eth_device_skip': {
                    'type': 'list',
                    'default': []
                },
            },
            'default': {},
        },
        'FIREWALL': {
            'type': 'dict',
            'default': {},
            'schema': {
                'port_blocking_mode': {
                    'type': 'string',
                    'default': PORT_BLOCKING_MODE_ALLOW,
                    'allowed': [PORT_BLOCKING_MODE_DENY,
                                PORT_BLOCKING_MODE_ALLOW],
                },
                'TCP_IN_IPv4': list_of_ports('default_ports_tpc_in_v4'),
                'TCP_OUT_IPv4': list_of_ports('default_ports_tcp_out_v4'),
                'UDP_IN_IPv4': list_of_ports('default_ports_udp_in_v4'),
                'UDP_OUT_IPv4': list_of_ports('default_ports_udp_out_v4'),
                # artifact from DEF-17989
                "internal_use_remote_iplist": {
                    "type": "boolean",
                    "default": False,
                },
            },
        },
        'DOS': {
            'type': 'dict',
            'schema': {
                'default_limit': {
                    'type': 'integer',
                    'coerce': (int, not_less_than(_DOS_DETECTOR_MIN_LIMIT)),
                    'min': _DOS_DETECTOR_MIN_LIMIT,
                    'default': _DOS_DETECTOR_DEFAULT_LIMIT,
                },
                'interval': {
                    'type': 'integer',
                    'coerce': (int, not_less_than(_DOS_DETECTOR_MIN_INTERVAL)),
                    'min': 1,
                    'default': 30,
                },
                'port_limits': {
                    'type': 'dict',
                    'keyschema': {
                        'type': 'integer',
                        'coerce': int,
                        'min': 1,
                        'max': 65535,
                    },
                    'valueschema': {
                        'type': 'integer',
                        'coerce': (int,
                                   not_less_than(_DOS_DETECTOR_MIN_LIMIT)),
                        'min': _DOS_DETECTOR_MIN_LIMIT,
                    },
                    'default': {},
                },
                'enabled': {
                    'type': 'boolean',
                    'default': True,
                }
            },
            'default': {},
        },
        'MOD_SEC': {
            'type': 'dict',
            'schema': {
                'app_specific_ruleset': {
                    'type': 'boolean',
                    'default': True,
                },
                'ruleset': {
                    'type': 'string',
                    'default': MODSEC_RULESET_FULL,
                    'allowed': [MODSEC_RULESET_MINIMAL, MODSEC_RULESET_FULL],
                },
                'cms_account_compromise_prevention': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
                'prev_settings': {
                    'type': 'string',
                    'required': False,
                    'default': '',
                },
            },
            'default': {},
        },
        'MOD_SEC_BLOCK_BY_SEVERITY': {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': True,
                },
                'max_incidents': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 2,
                },
                'check_period': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 120,
                },
                'severity_limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'max': 7,
                    'default': 2,
                },
                'denied_num_limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 2,
                }
            },
            'default': {},
        },
        'MOD_SEC_BLOCK_BY_CUSTOM_RULE': {
            'type': 'dict',
            'keyschema': {
                'type': 'integer',
                'coerce': int,
            },
            'valueschema': {
                'type': 'dict',
                'schema': {
                    'max_incidents': {
                        'type': 'integer',
                        'coerce': int,
                        'min': 1,
                        'default': 2,
                    },
                    'check_period': {
                        'type': 'integer',
                        'coerce': int,
                        'min': 1,
                        'default': 120,
                    },
                }
            },
            'default': {
                33332: {
                    'check_period': 120,
                    'max_incidents': 10
                },
                33339: {
                    'check_period': 120,
                    'max_incidents': 10
                }
            }
        },
        'INCIDENT_LOGGING': {
            'type': 'dict',
            'schema': {
                'num_days': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 100,
                },
                'limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 100000,
                },
                'min_log_level': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'default': 4,
                },
                'ui_autorefresh_timeout': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 5,
                    'default': 10,
                },
            },
            'default': {},
        },
        'MALWARE_CLEANUP': {
            'type': 'dict',
            'schema': {
                'trim_file_instead_of_removal': {
                    'type': 'boolean',
                    'default': True,
                },
                'keep_original_files_days': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'default': 14,
                },
            },
            'default': {},
        },
        'MALWARE_SCANNING': {
            'type': 'dict',
            'schema': {
                'try_restore_from_backup_first': {
                    'type': 'boolean',
                    'default': False,
                },
                'default_action': {
                    'type': 'string',
                    'default': CLEANUP,
                    'allowed': [NOTIFY, CLEANUP],
                },
                'notify_on_detect': {
                    'type': 'boolean',
                    'default': False,
                },
                'enable_scan_inotify': {
                    'type': 'boolean',
                    'default': True,
                },
                'optimize_realtime_scan': {
                    'type': 'boolean',
                    'default': True,
                },
                'enable_scan_modsec': {
                    'type': 'boolean',
                    'default': True,
                },
                'sends_file_for_analysis': {
                    'type': 'boolean',
                    'default': True,
                },
                'enable_scan_pure_ftpd': {
                    'type': 'boolean',
                    'default': True,
                },
                'scan_modified_files': {
                    'type': 'boolean',
                    'default': None,
                    'nullable': True,
                },
                'cloud_assisted_scan': {
                    'type': 'boolean',
                    'default': True,
                },
                'rapid_scan': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                "rapid_scan_rescan_unchanging_files_frequency": {
                    "type": "integer",
                    "coerce": nullable(int),
                    "nullable": True,
                    "default": None,
                },
                'max_signature_size_to_scan': {
                    'type': 'integer',
                    'coerce': int,
                    'default': 1048576,
                },
                'max_cloudscan_size_to_scan': {
                    'type': 'integer',
                    'coerce': int,
                    'default': 10485760,
                },
                'max_mrs_upload_file': {
                    'type': 'integer',
                    'coerce': int,
                    'default': 10485760,
                },
                'detect_elf': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                'hyperscan': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
                'enable_scan_cpanel': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
            },
            'default': {},
        },
        'MALWARE_SCAN_SCHEDULE': {
            'type': 'dict',
            'schema': {
                'interval': {
                    'type': 'string',
                    'allowed': [NONE, DAY, WEEK, MONTH],
                    'default': WEEK,
                },
                'hour': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'max': 23,
                    'default': 3,
                },
                'day_of_week': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 0,
                    'max': 7,
                    'default': 0,
                },
                'day_of_month': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 31,
                    'default': 1,
                },
            },
            'default': {},
        },
        'MALWARE_SCAN_INTENSITY': {
            'type': 'dict',
            'schema': {
                'cpu': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': DEFAULT_INTENSITY_CPU,
                },
                'io': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': DEFAULT_INTENSITY_IO,
                },
                'ram': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1024,
                    'default': DEFAULT_INTENSITY_RAM,
                },
                'user_scan_cpu': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': 2,
                },
                'user_scan_io': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': 2,
                },
                'user_scan_ram': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1024,
                    'default': 1024,
                },
            },
            'default': {},
        },
        'MALWARE_DATABASE_SCAN': {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': False,
                },
            },
            'default': {},
        },
        'RESOURCE_MANAGEMENT': {
            'type': 'dict',
            'schema': {
                'cpu_limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': DEFAULT_RESOURCE_MANAGEMENT_CPU_LIMIT,
                },
                'io_limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 1,
                    'max': 7,
                    'default': DEFAULT_RESOURCE_MANAGEMENT_IO_LIMIT,
                },
                'ram_limit': {
                    'type': 'integer',
                    'coerce': int,
                    'min': 500,
                    'default': DEFAULT_RESOURCE_MANAGEMENT_RAM_LIMIT,
                },
            },
            'default': {},
        },
        'WEB_SERVICES': {
            'type': 'dict',
            'schema': {
                'http_ports': {
                    'type': 'list',
                    'schema': {
                        'type': 'integer'
                    },
                    'default': []
                },
                'https_ports': {
                    'type': 'list',
                    'schema': {
                        'type': 'integer'
                    },
                    'default': []
                },
            },
            'default': {},
        },
        'CAPTCHA': {
            'type': 'dict',
            'schema': {
                'cert_refresh_timeout': {
                    'type': 'integer',
                    'default': 3600
                }
            },
            'default': {},
        },
        'ERROR_REPORTING': {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': True
                }
            },
            'default': {}
        },
        "CAPTCHA_DOS": {
            "type": 'dict',
            'schema': {
                'enabled': {
                    'type': 'boolean',
                    'default': True,
                },
                'time_frame': {
                    'type': 'integer',
                    'default': 21600,
                },
                'max_count': {
                    'type': 'integer',
                    'default': 100
                },
                'timeout': {
                    'type': 'integer',
                    'default': 864000
                }
            },
            'default': {},
        },
        "BLOCKED_PORTS": {
            'type': 'dict',
            'schema': {
                'default_mode': {
                    'type': 'string',
                    'default': 'allowed',
                    'allowed': ['allowed', 'denied']
                }
            },
            'default': {},
        },
        "SEND_ADDITIONAL_DATA": {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': True,
                }
            },
            'default': {}
        },
        "STOP_MANAGING": {
            "type": "dict",
            "schema": {
                "modsec_directives": {
                    "type": "boolean",
                    "default": False,
                },
            },
            "default": {},
        },
        "PROACTIVE_DEFENCE": {
            'type': 'dict',
            'coerce': (dict, proactive_defense_rules),
            'schema': {
                'mode': {
                    'type': 'string',
                    'nullable': False,
                    'default': 'LOG',
                    'allowed': ['DISABLED', 'LOG', 'KILL']
                },
                'blamer': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                'php_immunity': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
            },
            'default': {}
        },
        "WEBSHIELD": {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': True
                },
                'known_proxies_support': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True
                },
                'captcha_site_key': {
                    'type': 'string',
                    'default': '',
                },
                'captcha_secret_key': {
                    'type': 'string',
                    'default': '',
                },
                'splash_screen': {
                    'type': 'boolean',
                    'default': True,
                },
                'invisible_captcha': {
                    'type': 'boolean',
                    'default': False,
                },
            },
            'default': {}
        },
        "OSSEC": {
            'type': 'dict',
            'schema': {
                'active_response': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False
                }
            },
            'default': {}
        },
        'ADMIN_CONTACTS': {
            'type': 'dict',
            'schema': {
                'emails': {
                    'type': 'list',
                    'schema': {
                        'type': 'string'
                    },
                    'nullable': False,
                    'default': []
                },
                'enable_icontact_notifications': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True
                }
            },
            'default': {}
        },
        'PERMISSIONS': {
            'type': 'dict',
            'schema': {
                'support_form': {
                    'type': 'boolean',
                    'nullable': True,
                    'default': True,
                },
                'advisor': {
                    'type': 'boolean',
                    'nullable': True,
                    'default': True,
                },
                'user_ignore_list': {
                    'type': 'boolean',
                    'nullable': True,
                    'default': False,
                },
                'upgrade_button': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                'allow_malware_scan': {
                    'type': 'boolean',
                    'nullable': True,
                    'default': False,
                },
                'user_override_malware_actions': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
                'user_override_proactive_defense': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
            },
            'default': {},
        },
        'CSF_INTEGRATION': {
            'type': 'dict',
            'schema': {
                'catch_lfd_events': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                }
            },
            'default': {}
        },
        'PAM': {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                'exim_dovecot_protection': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': True,
                },
                'ftp_protection': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                }
            },
            'default': {}
        },
        'KERNELCARE': {
            'type': 'dict',
            'schema': {
                'edf': {
                    'type': 'boolean',
                    'nullable': False,
                    'default': False,
                },
            },
            'default': {}
        },
        'SMTP_BLOCKING': {
            'type': 'dict',
            'schema': {
                'enable': {
                    'type': 'boolean',
                    'default': False,
                },
                'ports': {
                    'type': 'list',
                    'schema': {
                        'type': 'integer',
                        'coerce': int,
                    },
                    'default': [25, 587, 465],
                },
                'allow_users': {
                    'type': 'list',
                    'schema': {
                        'type': 'string',
                    },
                    'default': [],
                },
                'allow_groups': {
                    'type': 'list',
                    'schema': {
                        'type': 'string',
                    },
                    'default': ['mail'],
                },
                'allow_local': {
                    'type': 'boolean',
                    'default': False,
                },
                'redirect': {
                    'type': 'boolean',
                    'default': False,
                }
            },
            'default': {},
        },
        'LOGGER': {
            'type': 'dict',
            'schema': {
                'max_log_file_size': {
                    'type': 'integer',
                    'default': 62914560,  # 60MB
                },
                'backup_count': {
                    # if 5, store app.log, app.log.1, up to app.log.5
                    'type': 'integer',
                    'default': 5,
                },
                'syscall_monitor': {
                    'type': 'boolean',
                    'default': False
                }
            },
            'default': {},
        }
    }


CONFIG_SCHEMA_NON_ROOT = {
    'BACKUP_RESTORE': {
        'type': 'dict',
        'schema': {
            'max_days_in_backup': {
                'type': 'integer',
                'nullable': True,
                'coerce': lambda x: x and int(x),
                'min': 1,
                'default': None,
            },
        },
        'default': {}
    },
    'MALWARE_SCANNING': {
        'type': 'dict',
        'schema': {
            'default_action': {
                'type': 'string',
                'default': None,
                'allowed': [NOTIFY, CLEANUP],
                'nullable': True
            },
            'try_restore_from_backup_first': {
                'type': 'boolean',
                'nullable': True,
                'default': None,
            }
        },
        'default': {},
    },
    "PROACTIVE_DEFENCE": {
        'type': 'dict',
        'schema': {
            'mode': {
                'type': 'string',
                'nullable': True,
                'allowed': ['DISABLED', 'LOG', 'KILL'],
                'default': None
            },
            'blamer': {
                'type': 'boolean',
                'nullable': True,
                'default': None,
            },
        },
        'default': {}
    },
}


CONFIG_SCHEMA_CUSTOM_BILLING = {
    'CUSTOM_BILLING': {
        'type': 'dict',
        'schema': {
            'upgrade_url': {
                'type': 'string',
                'default': None,
                'nullable': True,
            },
            "upgrade_url_360": {
                "type": "string",
                "default": None,
                "nullable": True,
            },
            "billing_notifications": {"type": "boolean", "default": True},
            "ip_license": {"type": "boolean", "default": True},
        },
        'default': {},
    }
}


class ConfigValidationError(Exception):
    pass


class Core:
    PRODUCT = 'imunify360'
    NAME = '%s agent' % PRODUCT if not ANTIVIRUS_MODE else 'imunify antivirus'
    VERSION = "6.6.0-2"
    API_BASE_URL = os.environ.get(
        "IMUNIFY360_API_URL", "https://api.imunify360.com"
    )
    DEFAULT_SOCKET_TIMEOUT = 10
    DIST = "xenial"
    FILE_UMASK = 0o007
    TMPDIR = "/var/imunify360/tmp"
    MERGED_CONFIG_FILE_NAME = "imunify360-merged.config"
    USER_CONFIG_FILE_NAME = "imunify360.config"
    LOCAL_CONFIG_FILE_NAME = "imunify360.config"
    CONFIG_DIR = "/etc/imunify360"
    USER_CONFDIR = os.path.join(CONFIG_DIR, "user_config")
    GLOBAL_CONFDIR = "/etc/sysconfig/imunify360"
    MERGED_CONFIG_FILE_PATH = os.path.join(
        GLOBAL_CONFDIR, MERGED_CONFIG_FILE_NAME
    )
    MERGED_CONFIG_FILE_PERMISSION = 0o644
    LOCAL_CONFIG_FILE_PATH = os.path.join(GLOBAL_CONFDIR, "imunify360.config")
    CONFIG_D_NAME = "imunify360.config.d"
    BACKUP_CONFIGFILENAME = '.imunify360.backup_config'
    HOOKS_CONFIGFILENAME = 'hooks.yaml'
    CUSTOM_BILLING_CONFIGFILENAME = 'custom_billing.config'
    INBOX_HOOKS_DIR = '/var/imunify360/hooks'

    SVC_NAME = 'imunify360' if not ANTIVIRUS_MODE else 'imunify-antivirus'
    UNIFIED_ACCESS_LOGGER_CONFIGFILENAME = "unified-access-logger.conf"


class IConfig(Protocol):
    @abstractmethod
    def config_to_dict(
        self, normalize: bool = True, force_read: bool = False
    ) -> dict:
        raise NotImplementedError

    @abstractmethod
    def dict_to_config(
        self,
        data: Mapping,
        validate: bool = True,
        normalize: bool = True,
        overwrite: bool = False,
        without_defaults: bool = False,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def validate(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def normalize(
        self, config: Mapping, without_defaults: bool = False
    ) -> dict:
        raise NotImplementedError

    @abstractmethod
    def modified_since(self, timestamp: Optional[float]) -> bool:
        raise NotImplementedError

    def get(self, section: str, option: Optional[str]) -> Any:
        if option:
            return self.config_to_dict().get(section, {}).get(option)
        return None

    def set(self, section: str, option: Optional[str], value: Any) -> None:
        if option:
            self.dict_to_config({section: {option: value}})


class IConfigFile(IConfig, Protocol):
    path: str


class ProtocolSingleton(_ProtocolMeta, Singleton):
    """
    Needed to avoid metaclass conflict when implementing protocols that are
    also Singletons
    """


class Normalizer:
    def __init__(self, validation_schema):
        self._config: Optional[Mapping] = None
        self._normalized_config: dict = {}
        self._schema = ConfigsValidator.get_validation_schema(
            validation_schema
        )
        self._schema_without_defaults = self._get_schema_without_defaults(
            self._schema
        )

    @classmethod
    def _get_schema_without_defaults(cls, _dict: Mapping) -> dict:
        return {
            key: cls._get_schema_without_defaults(value)
            if isinstance(value, dict)
            else value
            for key, value in _dict.items()
            if key != "default"
        }

    @staticmethod
    def remove_null(config: Mapping) -> dict:
        new_config: Dict[str, Union[dict, List[Tuple]]] = {}
        for section, options in config.items():
            # We only support dict for option removal
            if isinstance(options, dict):
                for option, value in options.items():
                    if value is not None:
                        new_config.setdefault(section, {})[option] = value
            elif options:
                # Let cerberus coerce this to dict
                # and leave the None's as is
                new_config[section] = options
        return new_config

    @staticmethod
    def _normalize_with_schema(config: Mapping, schema) -> dict:
        validator = ConfigValidator(schema)
        normalized: Optional[dict] = validator.normalized(config)
        if validator.errors:
            raise ConfigValidationError(validator.errors)
        if normalized is None:
            raise ConfigValidationError(f"Cerberus returned None for {config}")
        return normalized

    def normalize(self, config: Mapping, without_defaults: bool) -> dict:
        schema = (
            self._schema_without_defaults if without_defaults else self._schema
        )
        if without_defaults:
            config = self.remove_null(config)
            return self._normalize_with_schema(config, schema)
        # Utilize cache
        if config == self._config:
            return self._normalized_config
        normalized = self._normalize_with_schema(config, schema)
        self._config = config
        self._normalized_config = normalized
        return self._normalized_config


class Config(IConfig, metaclass=ProtocolSingleton):
    DISCLAIMER = ""

    def __init__(
        self,
        *,
        # FIXME: allow pathlib.Path
        path: str = None,
        config_reader: ConfigReader = None,
        validation_schema: Union[Mapping, Callable[[], Mapping]] = None,
        disclaimer: str = None,
        cached: bool = True,
        permissions: int = None,
    ):
        super().__init__()
        assert path or config_reader
        config_reader_cls = CachedConfigReader if cached else ConfigReader
        self._config_reader = config_reader or config_reader_cls(
            path, disclaimer=disclaimer or self.DISCLAIMER,
            permissions=permissions
        )
        self.path = self._config_reader.path
        self.validation_schema = validation_schema or config_schema_root
        self._normalizer = Normalizer(self.validation_schema)

    def __repr__(self):
        return (
            "<{classname}(config_reader={config_reader!r},"
            " validation_schema={validation_schema!r})"
            ">"
        ).format(
            classname=self.__class__.__qualname__,
            config_reader=self._config_reader,
            validation_schema=self.validation_schema,
        )

    def normalize(self, config: Mapping, without_defaults=False) -> dict:
        return self._normalizer.normalize(config, without_defaults)

    def config_to_dict(self, normalize=True, force_read: bool = False) -> dict:
        """
        Converts config file to dict

        :return dict: dictionary (key (section) / value (options))
        """
        config = self._config_reader.read_config_file(force_read=force_read)
        if normalize:
            config = self.normalize(config)

        return deepcopy(config)

    def dict_to_config(
        self,
        data: Mapping,
        validate: bool = True,
        normalize: bool = True,
        overwrite: bool = False,
        without_defaults: bool = False,
    ) -> None:
        """
        Converts dict to config file
        New options will be mixed in with old ones
        unless overwrite is specified

        :param dict data: dictionary (key (section) / value (options))
        :param bool validate: indicates if we need validation
        :param bool normalize: normalize config
        :param overwrite: overwrite existing conf
        :param without_defaults: do not fill defaults
        :return: None
        """
        if overwrite:
            self._dict_to_config_overwrite(
                data=data,
                validate=validate,
                normalize=normalize,
                without_defaults=without_defaults,
            )
        else:
            self._dict_to_config(
                data=data,
                validate=validate,
                normalize=normalize,
                without_defaults=without_defaults,
            )

    def _dict_to_config_overwrite(
        self, data, validate, normalize, without_defaults
    ):
        if validate:
            ConfigsValidator.validate(data, self.validation_schema)
        if normalize:
            data = self.normalize(data, without_defaults=without_defaults)
        self._config_reader.write_config_file(data)

    def _dict_to_config(self, data, validate, normalize, without_defaults):
        config = deepcopy(self._config_reader.read_config_file())
        if dict_deep_update(config, data):
            if validate:
                ConfigsValidator.validate(config, self.validation_schema)
            if normalize:
                config = self.normalize(
                    config, without_defaults=without_defaults
                )
            self._config_reader.write_config_file(config)

    def validate(self):
        """
        :raises ConfigsValidatorError
        """
        try:
            config_dict = self._config_reader.read_config_file(
                ignore_errors=False
            )
        except ConfigError as e:
            message = "Error during config validation"
            logger.exception("%s: %s", message, e)
            raise ConfigsValidatorError({self: message}) from e

        try:
            ConfigsValidator.validate(config_dict, self.validation_schema)
        except ConfigValidationError as e:
            message = "Imunify360 config does not match the scheme"
            logger.exception("%s: %s", message, e)
            raise ConfigsValidatorError({self: message}) from e

    def modified_since(self, timestamp: Optional[float]) -> bool:
        return self._config_reader.modified_since(timestamp)


class UserConfig(Config):
    def __init__(self, *, username):
        self.username = username
        path = os.path.join(
            Core.USER_CONFDIR, username, Core.USER_CONFIG_FILE_NAME
        )
        super().__init__(
            path=path,
            config_reader=UserConfigReader(path, username),
            validation_schema=CONFIG_SCHEMA_NON_ROOT,
        )


class SystemConfig(IConfig, metaclass=ProtocolSingleton):
    def __init__(
        self, *, local_config: Config = None, merged_config: Config = None
    ):
        super().__init__()
        self._local_config = local_config or LocalConfig()
        self._merged_config = merged_config or MergedConfig()

    def config_to_dict(
        self, normalize: bool = True, force_read: bool = False
    ) -> dict:
        return self._merged_config.config_to_dict(
            normalize=normalize, force_read=force_read
        )

    def dict_to_config(
        self,
        data: Mapping,
        validate: bool = True,
        normalize: bool = True,
        overwrite: bool = False,
        without_defaults: bool = True,
    ) -> None:
        self._local_config.dict_to_config(
            data,
            validate=validate,
            normalize=normalize,
            overwrite=overwrite,
            without_defaults=without_defaults,
        )

    def validate(self) -> None:
        self._merged_config.validate()

    def normalize(
        self, config: Mapping, without_defaults: bool = False
    ) -> dict:
        return self._merged_config.normalize(
            config=config, without_defaults=without_defaults
        )

    def modified_since(self, timestamp: Optional[float]) -> bool:
        return self._merged_config.modified_since(timestamp)


def config_file_factory(
    username: Optional[Union[str, int]] = None, path: Optional[str] = None
) -> IConfig:
    if username and not isinstance(username, int):
        return UserConfig(username=username)
    elif path:
        return Config(path=path)
    else:
        return SystemConfig()


# TODO: move all layer related functions to another module
def any_layer_modified_since(timestamp) -> bool:
    merger = Merger(Merger.get_layer_names())
    for layer in merger.layers:
        if layer.modified_since(timestamp):
            return True
    return False


MergedConfig = functools.partial(
    Config,
    path=Core.MERGED_CONFIG_FILE_PATH,
    disclaimer=DO_NOT_MODIFY_DISCLAMER,
    permissions=Core.MERGED_CONFIG_FILE_PERMISSION,
)


class LocalConfig(Config):
    """
    Config (/etc/sysconfig/imunify360/imunify360.config) should contain
    options changed by a customer only
    """
    def __init__(
        self,
        *,
        path=Core.LOCAL_CONFIG_FILE_PATH,  # overrides the parent default value
        config_reader: ConfigReader = None,
        validation_schema: Union[Mapping, Callable[[], Mapping]] = None,
        disclaimer: str = None,
        cached=False,  # overrides the parent default value
    ):
        super().__init__(
            path=path,
            config_reader=config_reader,
            validation_schema=validation_schema,
            disclaimer=disclaimer,
            cached=cached,
        )

    def dict_to_config(
        self,
        data: Mapping,
        validate: bool = True,
        normalize: bool = True,
        overwrite: bool = False,
        without_defaults: bool = True  # overrides the parent default value
    ) -> None:
        return super().dict_to_config(
            data,
            validate=validate,
            normalize=normalize,
            overwrite=overwrite,
            without_defaults=without_defaults
        )


class BaseMerger:
    DIR = os.path.join(Core.GLOBAL_CONFDIR, Core.CONFIG_D_NAME)
    LOCAL_CONFIG_NAME = "90-local.config"

    def __init__(self, names, include_defaults=False):
        self._include_defaults = include_defaults
        self.layers = [
            Config(path=os.path.join(self.DIR, name)) for name in names
        ]

    @classmethod
    def get_layer_names(cls):
        return sorted(os.listdir(cls.DIR)) if os.path.isdir(cls.DIR) else []

    def configs_to_dict(self, force_read=False):
        layer_dict_list = []
        if self._include_defaults:
            defaults = Normalizer(config_schema_root).normalize(
                {}, without_defaults=False
            )
            layer_dict_list.append(defaults)

        layer_dict_list += [
            layer.config_to_dict(normalize=False, force_read=force_read)
            for layer in self.layers
        ]
        return self._build_effective_config(layer_dict_list)

    @classmethod
    def _build_effective_config(cls, layer_dict_list: list):
        effective: Dict[str, dict] = {}
        for layer_dict in layer_dict_list:
            for section, options in layer_dict.items():
                if options is None:
                    continue
                if section not in effective:
                    effective[section] = {}
                for option, value in options.items():
                    if value is not None:
                        effective[section][option] = value
        return effective


class MutableMerger(BaseMerger):
    def __init__(self, names: Sequence):
        idx = bisect_left(names, self.LOCAL_CONFIG_NAME)
        names = names[:idx]
        super().__init__(names, include_defaults=True)


class ImmutableMerger(BaseMerger):
    def __init__(self, names: Sequence):
        idx = bisect_right(names, self.LOCAL_CONFIG_NAME)
        names = names[idx:]
        super().__init__(names, include_defaults=False)


class NonBaseMerger(BaseMerger):
    def __init__(self, names: Sequence):
        super().__init__(names, include_defaults=False)


class Merger(BaseMerger):
    def __init__(self, names):
        super().__init__(names, include_defaults=True)

    @classmethod
    def update_merged_config(cls):
        merged_config = MergedConfig()
        merger = cls(cls.get_layer_names())
        config_dict = merger.configs_to_dict()
        try:
            ConfigsValidator.validate(config_dict)
        except ConfigsValidatorError:
            logger.warning("Config file is invalid!")
        else:
            merged_config.dict_to_config(config_dict, validate=False)


class ConfigValidator(Validator):
    def _validate_is_port_or_range(self, is_port_or_range, field, value):
        """
        {'type': 'boolean'}
        """
        if not is_port_or_range:
            return

        def in_range(port):
            return 1 <= int(port) <= 65535

        ports = re.match(r'^(\d+)([:|-](\d+))?$', value)
        if not ports:
            self._error(field, 'Port should be integer or range (2-4 or 2:4)')
            return
        if not in_range(ports.group(1)):
            self._error(field, 'Port should be within range 0-65535')
            return
        if ports.group(3) is not None and not in_range(ports.group(3)):
            self._error(field, 'Port should be within range 0-65535')
            return

    def _normalize_coerce_sort_ports(self, value):
        return sorted(list(set([v.replace(':', '-') for v in value])),
                      key=lambda v: int(v.rsplit('-')[0]))

    def _normalize_default_setter_default_ports_tpc_in_v4(self, d):
        return self.get_default_ports('tcp', 'in')

    def _normalize_default_setter_default_ports_tcp_out_v4(self, d):
        return self.get_default_ports('tcp', 'out')

    def _normalize_default_setter_default_ports_udp_in_v4(self, d):
        return self.get_default_ports('udp', 'in')

    def _normalize_default_setter_default_ports_udp_out_v4(self, d):
        return self.get_default_ports('udp', 'out')

    @staticmethod
    def get_default_ports(protocol, direction):
        # to avoid circular dependency: HostingPanel loads ConfigReader
        from defence360agent.subsys.panels.hosting_panel import HostingPanel
        return HostingPanel().OPEN_PORTS[protocol][direction]


class Packaging:
    DATADIR = '/opt/alt/python38/share/%s' % Core.PRODUCT


class SimpleRpc:
    # how long to wait for the response from RPC server in seconds
    CLIENT_TIMEOUT = 3600
    SOCKET_PATH = '/var/run/defence360agent/simple_rpc.sock'
    # end-user stuff
    NON_ROOT_SOCKET_PATH = '/var/run/defence360agent/non_root_simple_rpc.sock'
    TOKEN_FILE_TMPL = '.imunify360_token_{suffix}'
    # -r--------
    TOKEN_MASK = 0o400
    SOCKET_ACTIVATION = bool_from_envvar(
        "I360_SOCKET_ACTIVATION",
        ANTIVIRUS_MODE,
    )
    INACTIVITY_TIMEOUT = int_from_envvar(
        "IMUNIFY360_INACTIVITY_TIMEOUT",
        int(timedelta(minutes=5).total_seconds()),
    )


class Model:
    #   type  sqlite3 - sqlite only supported
    PATH = '/var/%s/%s.db' % (Core.PRODUCT, Core.PRODUCT)
    PROACTIVE_PATH = '/var/%s/proactive.db' % (Core.PRODUCT, )


class FilesUpdate:
    # Check files update periodically
    PERIOD = timedelta(minutes=30).total_seconds()
    # Timeout for single Index update (group of files) DEF-11501
    # It has to be less than watchdog timeout (60 min)
    TIMEOUT = timedelta(minutes=15).total_seconds()
    # Timeout for individual socket operations (connect, recv/read, etc.)
    # For instance ssl-handshake timeout == SOCKET_TIMEOUT
    # Than less the value than better, to do not wait to long
    SOCKET_TIMEOUT = 3  # seconds


class CountryInfo:
    DB = '/var/imunify360/files/geo/v1/GeoLite2-Country.mmdb'

    LOCATIONS_DB = '/var/imunify360/files/geo/v1/' \
                   'GeoLite2-Country-Locations-en.csv'

    @staticmethod
    def country_subnets_file(country_code):
        return '/var/imunify360/files/geo/v1/CountrySubnets-{}.txt'.format(
            country_code)


class Sentry:
    DSN = os.getenv('IMUNITY360_SENTRY_DSN',
                    _slurp_file('%s/sentry' % Packaging.DATADIR))
    ENABLE = FromConfig('ERROR_REPORTING', 'enable')


class Malware:
    SCAN_CHECK_PERIOD = 300
    CONSECUTIVE_ERROR_LIMIT = 10
    INOTIFY_SCAN_PERIOD = 60
    CONFIG_CHECK_PERIOD = 30
    CONFLICTS_CHECK_PERIOD = 300
    INOTIFY_ENABLED = FromConfig('MALWARE_SCANNING', 'enable_scan_inotify')
    PURE_SCAN = FromConfig('MALWARE_SCANNING', 'enable_scan_pure_ftpd')

    SEND_FILES = FromConfig('MALWARE_SCANNING', 'sends_file_for_analysis')
    CLOUD_ASSISTED_SCAN = FromConfig('MALWARE_SCANNING', 'cloud_assisted_scan')
    RAPID_SCAN = FromConfig('MALWARE_SCANNING', 'rapid_scan')

    SCANS_PATH = '/var/imunify360/aibolit/scans.pickle'

    FILE_PREVIEW_BYTES_NUM = 1024 * 100  # 100 KB

    CLEANUP_STORAGE = '/var/imunify360/cleanup_storage'
    CLEANUP_TRIM = FromConfig(
        section='MALWARE_CLEANUP',
        option='trim_file_instead_of_removal',
    )
    CLEANUP_KEEP = FromConfig(
        section='MALWARE_CLEANUP',
        option='keep_original_files_days',
    )
    SCAN_MODIFIED_FILES = FromConfig(
        section='MALWARE_SCANNING',
        option='scan_modified_files',
    )

    MAX_SIGNATURE_SIZE_TO_SCAN = FromConfig('MALWARE_SCANNING',
                                            'max_signature_size_to_scan')
    MAX_CLOUDSCAN_SIZE_TO_SCAN = FromConfig('MALWARE_SCANNING',
                                            'max_cloudscan_size_to_scan')
    MAX_MRS_UPLOAD_FILE = FromConfig('MALWARE_SCANNING', 'max_mrs_upload_file')

    USE_JSON_REPORT = FromFlagFile("use_json")
    RAPID_SCAN_RESCAN_UNCHANGING_FILES_FREQUENCY = FromConfig(
        "MALWARE_SCANNING", "rapid_scan_rescan_unchanging_files_frequency"
    )

    HYPERSCAN = FromConfig('MALWARE_SCANNING', 'hyperscan')

    DATABASE_SCAN_ENABLED = FromConfig("MALWARE_DATABASE_SCAN", "enable")
    CPANEL_SCAN_ENABLED = FromConfig('MALWARE_SCANNING', 'enable_scan_cpanel')


class MalwareGeneric:
    BASEDIR = None
    PATTERN_TO_WATCH = ''


# static initializer for MalwareGeneric class
integration_conf = IntegrationConfig()
if integration_conf.exists():
    try:
        malware_section = integration_conf.to_dict()['malware']
    except KeyError:
        pass
        # inotify settings are not necessary
        # it even may be not supported by system
    else:
        MalwareGeneric.BASEDIR = malware_section['basedir']
        MalwareGeneric.PATTERN_TO_WATCH\
            = malware_section['pattern_to_watch']


class MalwareScanScheduleInterval:
    NONE = NONE
    DAY = DAY
    WEEK = WEEK
    MONTH = MONTH


class MalwareScanSchedule:
    CMD = '/usr/bin/imunify360-agent malware user scan --background'
    CRON_PATH = '/etc/cron.d/imunify_scan_schedule'
    CRON_STRING = '''\
# DO NOT EDIT. AUTOMATICALLY GENERATED.
0 {0} {1} * {2} root {cmd} >/dev/null 2>&1
'''
    INTERVAL = FromConfig(
        section='MALWARE_SCAN_SCHEDULE',
        option='interval',
    )
    HOUR = FromConfig(
        section='MALWARE_SCAN_SCHEDULE',
        option='hour',
    )
    DAY_OF_WEEK = FromConfig(
        section='MALWARE_SCAN_SCHEDULE',
        option='day_of_week',
    )
    DAY_OF_MONTH = FromConfig(
        section='MALWARE_SCAN_SCHEDULE',
        option='day_of_month',
    )


class MalwareScanIntensity:
    CPU = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='cpu',
    )
    IO = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='io',
    )
    RAM = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='ram',
    )
    USER_CPU = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='user_scan_cpu',
    )
    USER_IO = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='user_scan_io',
    )
    USER_RAM = FromConfig(
        section='MALWARE_SCAN_INTENSITY',
        option='user_scan_ram',
    )


class FileBasedResourceLimits:
    CPU = FromConfig(
        section='RESOURCE_MANAGEMENT',
        option='cpu_limit',
    )
    IO = FromConfig(
        section='RESOURCE_MANAGEMENT',
        option='io_limit',
    )
    RAM = FromConfig(
        section='RESOURCE_MANAGEMENT',
        option='ram_limit',
    )


class KernelCare:
    EDF = FromConfig(
        section='KERNELCARE',
        option='edf',
    )


def choose_action_for_malicious(
    username: str,
) -> Tuple[str, str]:
    if PermissionsConfig.USER_OVERRIDE_MALWARE_ACTIONS:
        return choose_value_from_config(
            "MALWARE_SCANNING", "default_action", username
        )

    return choose_value_from_config("MALWARE_SCANNING", "default_action")


def get_rapid_rescan_frequency():
    value = Malware.RAPID_SCAN_RESCAN_UNCHANGING_FILES_FREQUENCY
    if value is None:
        freq = {
            MalwareScanScheduleInterval.NONE: 1,
            MalwareScanScheduleInterval.MONTH: 2,
            MalwareScanScheduleInterval.WEEK: 5,
            MalwareScanScheduleInterval.DAY: 10,
        }
        return freq.get(MalwareScanSchedule.INTERVAL, 1)
    return value


class MalwareSignatures:
    _dir = '/var/imunify360/files/sigs/v1/'
    RFXN = os.path.join(_dir, 'rfxn')
    i360 = os.path.join(_dir, 'i360')

    AI_BOLIT_HOSTER = os.path.join(_dir, "aibolit", "ai-bolit-hoster-full.db")
    AI_BOLIT_HYPERSCAN = os.path.join(_dir, "aibolit", "hyperscan")
    MDS_AI_BOLIT_HOSTER = os.path.join(
        _dir, "aibolit", "mds-ai-bolit-hoster.db"
    )
    PROCU_DB = os.path.join(_dir, "aibolit", "procu2.db")
    MDS_PROCU_DB = os.path.join(_dir, "aibolit", "mds-procu2.db")


class Logger:
    MAX_LOG_FILE_SIZE = FromConfig(
        section='LOGGER',
        option='max_log_file_size',
    )
    BACKUP_COUNT = FromConfig(
        section='LOGGER',
        option='backup_count',
    )
    # directory mode for main log directory and all inner
    LOG_DIR_PERM = 0o700
    # file mode for main log directory and all inner
    LOG_FILE_PERM = 0o660


class UserType:
    ROOT = 'root'
    NON_ROOT = 'non_root'


class UIRole:
    CLIENT = 'client'
    ADMIN = 'admin'


class NoCP:
    # path to script implementing No CP API
    CLIENT_SCRIPT = '/etc/imunify360/scripts/domains'
    # latest version of the API supported by agent
    LATEST_VERSION = 1


class CustomBillingConfig(Config):
    def __init__(self):
        path = os.path.join(
            "/etc/sysconfig/imunify360", Core.CUSTOM_BILLING_CONFIGFILENAME
        )
        super().__init__(
            path=path, validation_schema=CONFIG_SCHEMA_CUSTOM_BILLING
        )


class CustomBilling:
    UPGRADE_URL = FromConfig(
        section='CUSTOM_BILLING',
        option='upgrade_url',
        config_cls=CustomBillingConfig
    )
    UPGRADE_URL_360 = FromConfig(
        section="CUSTOM_BILLING",
        option="upgrade_url_360",
        config_cls=CustomBillingConfig,
    )
    NOTIFICATIONS = FromConfig(
        section='CUSTOM_BILLING',
        option='billing_notifications',
        config_cls=CustomBillingConfig
    )
    IP_LICENSE = FromConfig(
        section='CUSTOM_BILLING',
        option='ip_license',
        config_cls=CustomBillingConfig
    )


class PermissionsConfig:
    USER_IGNORE_LIST = FromConfig(
        section='PERMISSIONS',
        option='user_ignore_list',
    )
    ALLOW_MALWARE_SCAN = FromConfig(
        section='PERMISSIONS',
        option='allow_malware_scan',
    )
    USER_OVERRIDE_MALWARE_ACTIONS = FromConfig(
        section='PERMISSIONS',
        option='user_override_malware_actions'
    )
    USER_OVERRIDE_PROACTIVE_DEFENSE = FromConfig(
        section='PERMISSIONS',
        option='user_override_proactive_defense',
    )


def effective_user_config(admin_config, user_config):
    allowed_sections = [
        "BACKUP_RESTORE",
        "MALWARE_CLEANUP",
        "MALWARE_SCANNING",
        "ERROR_REPORTING",
        "PROACTIVE_DEFENCE",
        "PERMISSIONS",
    ]
    overridable = {
        (
            "MALWARE_SCANNING",
            "default_action",
        ): PermissionsConfig.USER_OVERRIDE_MALWARE_ACTIONS,
        (
            "PROACTIVE_DEFENCE",
            "mode",
        ): PermissionsConfig.USER_OVERRIDE_PROACTIVE_DEFENSE,
    }
    admin_dict = admin_config.config_to_dict()
    user_dict = user_config.config_to_dict()

    def normalize_section(section):
        admin_options = deepcopy(admin_dict.get(section))
        user_options = deepcopy(user_dict.get(section, {}))
        resulting_dict = {}
        for option, admin_value in admin_options.items():
            user_value = user_options.get(option)
            if (
                user_value is not None
                # All options available to user are overridable by default
                and overridable.get((section, option), True)
            ):
                resulting_dict[option] = user_value
            else:
                resulting_dict[option] = admin_value

        return resulting_dict

    effective_config = {
        section: normalize_section(section) for section in allowed_sections
    }

    return fm_config_cleanup(effective_config, user_config.username)


class HookEvents:
    IM360_EVENTS = (
        AGENT, LICENSE, MALWARE_SCANNING,
        MALWARE_CLEANUP, MALWARE_DETECTED,
    ) = (
        'agent', 'license', 'malware-scanning',
        'malware-cleanup', 'malware-detected',
    )
    IMAV_EVENTS = (
        LICENSE, MALWARE_SCANNING, MALWARE_CLEANUP, MALWARE_DETECTED
    )
    EVENTS = IMAV_EVENTS if ANTIVIRUS_MODE else IM360_EVENTS


class ConfigsValidatorError(Exception):
    def __init__(self, configs_to_errors: Dict[Config, str]):
        self.configs_to_errors = configs_to_errors

    def __repr__(self):
        errors = []
        for config, error in self.configs_to_errors.items():
            errors.append(f"{config!r}: {error}")
        return "\n".join(errors)


class ConfigsValidator:
    """ A class that has methods to validate configs bypassing the cache """

    @classmethod
    def validate_system_config(cls):
        """
        Validate merged config
        :raises ConfigsValidatorError
        """
        SystemConfig().validate()

    @classmethod
    def validate_config_layers(cls):
        """
        Validate all config layers, collect all errors
        :raises ConfigsValidatorError
        """
        configs_to_errors = {}
        for layer in Merger(Merger.get_layer_names()).layers:
            try:
                layer.validate()
            except ConfigsValidatorError as e:
                configs_to_errors.update(e.configs_to_errors)

        if configs_to_errors:
            raise ConfigsValidatorError(configs_to_errors)

    @classmethod
    def validate(
        cls,
        config_dict: dict,
        validation_schema: Union[dict, Callable] = config_schema_root,
    ) -> None:
        """
        Validate config represented by a dict
        :param config_dict: config to validate
        :param validation_schema: schema to validate config against
        :raises ConfigValidationError
        """

        schema = cls.get_validation_schema(validation_schema)
        v = ConfigValidator(schema)
        if not v.validate(config_dict):
            raise ConfigValidationError(v.errors)

    @staticmethod
    def get_validation_schema(
        validation_schema: Union[Mapping, Callable]
    ) -> Mapping:
        if callable(validation_schema):
            return validation_schema()
        return validation_schema


ConfigFile = config_file_factory


class AdminContacts:
    ENABLE_ICONTACT_NOTIFICATIONS = FromConfig(
        section="ADMIN_CONTACTS",
        option="enable_icontact_notifications",
    )


class IContactMessageType(str, Enum):
    MALWARE_FOUND = "MalwareFound"
    SCAN_NOT_SCHEDULED = "ScanNotScheduled"
    GENERIC = "Generic"

    def __str__(self):
        return self.value
