import asyncio
import contextlib
import json
import os
from collections import defaultdict
from heapq import heappop, heappush
from logging import getLogger
from pathlib import Path
from typing import Dict, Optional

from defence360agent.contracts.config import ConfigFile, Core
from defence360agent.contracts.messages import cPanelEvent
from defence360agent.contracts.plugins import (
    BaseMessageProcessor,
    MessageSource,
    expect,
)
from defence360agent.feature_management.constants import (
    AV,
    EXTENSION_DEFAULTS,
    EXTENSION_NAME,
    FEATURE_EXT_VARIABLES,
    PROACTIVE,
)
from defence360agent.feature_management.model import FeatureManagementPerms
from defence360agent.feature_management.utils import set_feature
from defence360agent.model.simplification import run_in_executor
from defence360agent.subsys.panels import hosting_panel
from defence360agent.subsys.panels.cpanel import packages
from defence360agent.utils import recurring_check

logger = getLogger(__name__)


class NativeEventMonitor(MessageSource):
    EVENT_DIR = Core.INBOX_HOOKS_DIR
    PATTERN = "*.*.*.*.json"

    async def create_source(self, loop, sink):
        self._loop = loop
        self._sink = sink
        self._hp = hosting_panel.HostingPanel()
        self._processor = NativeSettingsChange(loop)
        self._processing_task = self._loop.create_task(
            self._check_inbox_folder_generate_events())

    async def shutdown(self):
        self._processing_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._processing_task

    @staticmethod
    def _rmfile(file: Path):
        try:
            file.unlink()
        except FileNotFoundError:
            pass  # do nothing if we cannot remove it, just skip it
        except Exception as e:
            logger.warning("Couldn't remove file %s %s", file, e)

    @staticmethod
    def _from_json(file: Path) -> Dict:
        return json.loads(file.read_text())

    def _event_to_message(self, file) -> Optional[cPanelEvent]:
        try:
            username, hook, ts1, ts2, *_ = file.name.split('.')
            ts = float(ts1 + '.' + ts2)
        except ValueError:
            logger.warning('hook-event-file detected with wrong name %s', file)
            return None
        try:
            return cPanelEvent.from_hook_event(
                username=username, hook=hook, ts=ts,
                fields=self._from_json(file))
        except FileNotFoundError:
            # already deleted
            logger.warning('hook file disappeared %s', file)
        except json.JSONDecodeError:
            # wrong format or broken json
            logger.warning('hook file have broken json %s', file)
        return None

    @recurring_check(30)
    async def _check_inbox_folder_generate_events(self):
        if not await self._hp.is_native_feature_management_enabled():
            return
        messages = []
        for file in Path(self.EVENT_DIR).glob('*.*.*.json'):
            try:
                message = self._event_to_message(file)
                if message is not None:
                    messages.append(message)
            except Exception as exc:
                logger.error('Failed to process %s hook event', exc)
            finally:
                self._rmfile(file)
        for message in messages:
            self._processor.add_message(message)
        await self._processor.process_messages()


class NativeSettingsChange(BaseMessageProcessor):
    """Proccess hook event messages from cPanel"""

    FEATURES_LIST = [PROACTIVE, AV]

    def __init__(self, loop):
        # note: empty list is a heap (no need for heapify here)
        self._msg_buf = defaultdict(list)
        self._loop = loop

    def add_message(self, message):
        heappush(
            self._msg_buf[message['username']],
            (message['timestamp'], message)
        )

    async def process_messages(self):
        await asyncio.gather(*(
            self.process_user_messages(user_messages)
            for user_messages in self._msg_buf.values()
        ))

    async def process_user_messages(self, messages):
        for _ in range(len(messages)):
            await self.process_message(heappop(messages)[1])

    @expect(cPanelEvent, hook='Modify')
    async def process_modify(self, message):
        if 'old_username' in message.data:  # User renamed
            old_username = message.data['old_username']
            await run_in_executor(
                self._loop,
                lambda: (
                    FeatureManagementPerms.update(user=message.username)
                    .where(FeatureManagementPerms.user == old_username)
                    .execute()
                ))
            os.rename(
                os.path.join(Core.USER_CONFDIR, old_username),
                os.path.join(Core.USER_CONFDIR, message.username)
            )
        package_field = 'plan' if 'plan' in message.data else 'exclude'
        await self._get_settings_and_update(message, package_field)

    @expect(cPanelEvent, hook='Create')
    async def process_create(self, message):
        await self._get_settings_and_update(message, 'plan', True)

    @expect(cPanelEvent, hook='change_package')
    async def process_change_package(self, message):
        await self._get_settings_and_update(message, 'new_pkg', True)

    @expect(cPanelEvent, hook='Remove')
    async def process_account_removed(self, message):
        user = message.get('user') or message.get('username')
        if user:
            logger.info('Resetting FM settings for %s', user)
            await run_in_executor(
                self._loop,
                lambda: (FeatureManagementPerms.delete()
                         .where(FeatureManagementPerms.user == user)
                         .execute()))
            try:
                os.unlink(ConfigFile(user).path)
            except FileNotFoundError:
                pass
            except OSError:
                logger.warning(
                    "Cannot delete Imunify360 config file for user %s", user
                )

    @staticmethod
    def _default_settings() -> Dict[str, str]:
        return {feature: EXTENSION_DEFAULTS[pe_var]
                for feature, pe_var in FEATURE_EXT_VARIABLES.items()}

    async def _get_settings_and_update(
        self,
        message,
        package_field: str,
        add_to_package: bool = False,
    ) -> None:
        logger.info("Get settings from %s", message)
        settings = {}
        for feature in self.FEATURES_LIST:
            try:
                value = message.data[FEATURE_EXT_VARIABLES[feature]]
            except KeyError:
                value = None
            settings[feature] = value
        logger.info("Step 1 %s ", settings)
        # Do nothing if there are no values for Imunify360 features
        # in the message for Modify hook
        if (
            message.get("plan") is None
            and message["hook"] == "Modify"
            and all(value is None for value in settings.values())
        ):
            return
        if not all(settings.values()):
            try:
                package_name = message.data[package_field]
            except KeyError:
                logger.warning("No information about package in message")
                fallback_settings = self._default_settings()
            else:
                fallback_settings = await self._get_package_settings(
                    package_name, add_to_package)
            for feature, value in settings.items():
                if value is None:
                    settings[feature] = fallback_settings[feature]
        logger.info("Settings specified in hook message %s for %s", settings,
                    message['username'])
        for feature, value in settings.items():
            await set_feature(message['username'], feature, value)

    @classmethod
    async def _get_package_settings(cls, package_name: str,
                                    add_to_package: bool) -> Dict[str, str]:
        logger.info('Getting package settings %s', package_name)
        try:
            pkg_info = await packages.get_package_info(package_name)
            return {feature: pkg_info[pe_var]
                    for feature, pe_var in FEATURE_EXT_VARIABLES.items()}
        except KeyError:
            logger.info("No extension's fields in package settings %s",
                        pkg_info)
            if add_to_package:
                await packages.add_extension(EXTENSION_NAME, pkg_info,
                                             **EXTENSION_DEFAULTS)
        except packages.PackageNotExistError:
            logger.warning("Package %s doesn't exist", package_name)

        return cls._default_settings()
