import asyncio
import contextlib
from logging import getLogger
from defence360agent.api import inactivity

from defence360agent.contracts.plugins import MessageSource
from defence360agent.feature_management.utils import sync_users
from defence360agent.malwarelib.subsys.ainotify import Inotify, Watcher
from defence360agent.subsys.panels import hosting_panel
from defence360agent.utils import finally_happened, recurring_check

logger = getLogger(__name__)


class PasswdWatcher(MessageSource):
    WATCH_PATH = b'/etc'
    WATCH_FILE = b'passwd'
    WATCH_MASK = Inotify.MOVED_TO

    async def create_source(self, loop, sink, watcher=None):
        self._loop = loop
        self._sink = sink

        self._hosting_panel = hosting_panel.HostingPanel()

        if watcher is None:
            self._watcher = Watcher(self._loop)
        else:
            self._watcher = watcher
        self._watcher.watch(self.WATCH_PATH, self.WATCH_MASK)

        if not await self._hosting_panel.is_native_feature_management_enabled():  # noqa: E501
            await sync_users(await self._hosting_panel.get_users())

        self._watch_task = self._loop.create_task(self._process_events())

    async def shutdown(self):
        self._watch_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._watch_task

        self._watcher and self._watcher.close()

    async def update_users_permissions(self) -> bool:
        return await sync_users(await self._hosting_panel.get_users())

    @recurring_check(0)
    async def _process_events(self):
        event = await self._watcher.get_event()
        if event.name == self.WATCH_FILE and \
                (not await self._hosting_panel.is_native_feature_management_enabled()):  # noqa: E501
            # updating the users list may take a while on different panels
            with inactivity.track.task("permissions_update"):
                await finally_happened(
                    self.update_users_permissions, max_tries=5, delay=1
                )
                logger.info('Feature management permissions updated')
