import asyncio
from contextlib import suppress
import grp
import json
import os
from pathlib import Path
import random
import time
from logging import getLogger
from urllib.parse import urljoin
from urllib.request import Request

from defence360agent.api.server import API, APIError
from defence360agent.contracts.license import LicenseCLN
from defence360agent.utils import atomic_rewrite
from defence360agent.utils.common import DAY

logger = getLogger(__name__)


_MAX_TRIES = 10
_TIMEOUT_MULTIPLICATOR = 0.5
"""
>>> _MAX_TRIES_FOR_DOWNLOAD = 10
>>> _TIMEOUT_MULTIPLICATOR = 0.5
>>> [(1 << i) * _TIMEOUT_MULTIPLICATOR for i in range(1, _MAX_TRIES_FOR_DOWNLOAD)]  # noqa
[1.0, 2.0, 4.0, 8.0, 16.0, 32.0, 64.0, 128.0, 256.0]
"""
_ACTIVATE_MINIMUM_TIMEOUT = 60


class IndependentAgentIDAPI(API):
    API_PATH = "/api/auth/agent/{}"
    REGISTER_URL = urljoin(API._BASE_URL, API_PATH.format("register"))
    ACTIVATE_URL = urljoin(API._BASE_URL, API_PATH.format("activate"))
    LOGIN_URL = urljoin(API._BASE_URL, API_PATH.format("login"))

    IAID_DIR = Path("/var/imunify360")
    IAID_FILE = IAID_DIR / "iaid"
    IAID_PASSWORD_FILE = IAID_DIR / "iaid-password"
    IAID_TOKEN_FILE = IAID_DIR / "iaid-token"
    IAID_ACTIVATED_FILE = IAID_DIR / "iaid-activated"
    _tasks = {
        "register": [],
        "activate": [],
        "login": [],
    }

    @staticmethod
    async def _retry_on_error(coro, attempt, timeout=0):
        # Exponential backoff retry
        await asyncio.sleep(
            timeout + random.randrange(1 << attempt) * _TIMEOUT_MULTIPLICATOR
        )
        await coro

    @classmethod
    def _add_task(cls, type, coro, attempt, timeout=0):
        cls._tasks[type] = [
            task for task in cls._tasks[type] if not task.done()
        ]
        if len(cls._tasks[type]) <= 1:
            loop = asyncio.get_event_loop()
            cls._tasks[type].append(
                loop.create_task(cls._retry_on_error(coro, attempt, timeout))
            )
        else:
            coro.close()
            logger.info("Task %s already in retry queue", type)

    @classmethod
    async def shutdown(cls):
        for type, tasks in cls._tasks.items():
            for task in tasks:
                if not task.done():
                    task.cancel()
                    with suppress(asyncio.CancelledError):
                        await task
                    logger.info("Retry task %s was canceled.", type)

    @staticmethod
    def _gid():
        return grp.getgrnam("_imunify").gr_gid

    @classmethod
    def get_iaid(cls):
        if cls.IAID_FILE.exists():
            return cls.IAID_FILE.read_text()
        return None

    @staticmethod
    def _request(url, **kwargs):
        return Request(
            url,
            method="POST",
            headers={"Content-Type": "application/json"},
            data=json.dumps(kwargs).encode() if kwargs else None,
        )

    @classmethod
    async def async_request(cls, request):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, cls.request, request)

    @classmethod
    def is_registered(cls):
        return all(
            iaid_file.exists()
            for iaid_file in (cls.IAID_FILE, cls.IAID_PASSWORD_FILE)
        )

    @classmethod
    def get_token(cls):
        try:
            return cls.IAID_TOKEN_FILE.read_text()
        except FileNotFoundError:
            return None

    @classmethod
    def is_token_expired(cls):
        try:
            stat = os.stat(cls.IAID_TOKEN_FILE)
        except FileNotFoundError:
            st_mtime = 0.0
        else:
            st_mtime = stat.st_mtime
        return time.time() - st_mtime > DAY

    @classmethod
    async def register(cls, force=False, attempt=1):
        if not force and cls.is_registered():
            return

        payload = dict()
        server_id = LicenseCLN.get_server_id()
        if server_id:
            payload["server_id"] = server_id

        request = cls._request(cls.REGISTER_URL, **payload)
        try:
            result = await cls.async_request(request)
            cls.IAID_ACTIVATED_FILE.unlink(missing_ok=True)
        except APIError as e:
            logger.error(
                "something went wrong on register %r - attempt %s", e, attempt
            )
            if (
                e.status_code >= 500 or e.status_code == 402
            ) and attempt < _MAX_TRIES:
                # internal error we may try again
                cls._add_task(
                    "register",
                    cls.register(force=force, attempt=attempt + 1),
                    attempt,
                )
                return
        else:
            atomic_rewrite(
                str(cls.IAID_FILE),
                result["iaid"],
                backup=cls.IAID_FILE.exists(),
                uid=-1,
                gid=cls._gid(),
                permissions=0o640,
            )
            atomic_rewrite(
                str(cls.IAID_PASSWORD_FILE),
                result["password"],
                backup=cls.IAID_PASSWORD_FILE.exists(),
                permissions=0o600,
            )
            await cls.activate()

    @classmethod
    async def activate(cls, attempt=1):
        if cls.IAID_ACTIVATED_FILE.exists():
            return
        if not cls.is_registered():
            logger.error("need to register first")
            await cls.register()
            return
        if LicenseCLN.is_free():
            if cls.is_token_expired():
                await cls.login()
            return
        iaid = cls.IAID_FILE.read_text()
        password = cls.IAID_PASSWORD_FILE.read_text()
        lic = LicenseCLN.get_token()

        request = cls._request(
            cls.ACTIVATE_URL, iaid=iaid, password=password, license=lic
        )
        try:
            await cls.async_request(request)
            cls.IAID_ACTIVATED_FILE.touch()
            await cls.login()
        except APIError as e:
            logger.error(
                "something went wrong on activate %r attempt %s", e, attempt
            )
            if e.status_code and e.status_code == 401:
                await cls.register(force=True)
            elif (
                e.status_code
                and (e.status_code >= 500 or e.status_code == 402)
                and attempt < _MAX_TRIES
            ):
                # internal error we may try again
                # 402 - if it is fresh registration it may took
                # time to sync CLN db
                cls._add_task(
                    "activate",
                    cls.activate(attempt=attempt + 1),
                    attempt,
                    timeout=_ACTIVATE_MINIMUM_TIMEOUT,
                )

    @classmethod
    async def reactivate(cls):
        cls.IAID_ACTIVATED_FILE.unlink(missing_ok=True)
        await cls.activate()

    @classmethod
    async def login(cls, attempt=1):
        if not cls.is_registered():
            logger.error("need to register first")
            return
        iaid = cls.IAID_FILE.read_text()
        password = cls.IAID_PASSWORD_FILE.read_text()

        request = cls._request(cls.LOGIN_URL, iaid=iaid, password=password)
        try:
            result = await cls.async_request(request)
        except APIError as e:
            logger.error(
                "something wrong happened on login %r attempt %s", e, attempt
            )
            if attempt < _MAX_TRIES:
                if e.status_code is None or e.status_code >= 500:
                    # internal error we may try again
                    cls._add_task(
                        "login", cls.login(attempt=attempt + 1), attempt
                    )
                elif e.status_code == 401:
                    await cls.register(force=True)
        else:
            atomic_rewrite(
                str(cls.IAID_TOKEN_FILE),
                result["token"],
                backup=cls.IAID_TOKEN_FILE.exists(),
                uid=-1,
                gid=cls._gid(),
                permissions=0o640,
            )
