import asyncio
import collections
import contextlib
import os
from logging import getLogger

from defence360agent.contracts.messages import Accumulatable
from defence360agent.contracts.plugins import (MessageSink, MessageSource,
                                               expect)
from defence360agent.utils import recurring_check

logger = getLogger(__name__)


class Accumulate(MessageSink, MessageSource):
    PROCESSING_ORDER = MessageSink.ProcessingOrder.POST_PROCESS_MESSAGE
    DEFAULT_AGGREGATE_TIMEOUT = int(
        os.environ.get("IMUNIFY360_AGGREGATE_MESSAGES_TIMEOUT", 60)
    )

    def __init__(self, period=DEFAULT_AGGREGATE_TIMEOUT, **kwargs):
        super().__init__(**kwargs)
        self._period = period

    async def create_source(self, loop, sink):
        self._loop = loop
        self._sink = sink
        self._task = loop.create_task(
            recurring_check(self._period)(self._flush)())
        self._data = collections.defaultdict(list)

    async def create_sink(self, loop):
        self._loop = loop

    async def shutdown(self):
        if self._task is not None:
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task

    @expect(Accumulatable)
    async def collect(self, message):
        list_type = message.LIST_CLASS
        if message.do_accumulate():
            self._data[list_type].append(message)

    async def _flush(self):
        copy_data = self._data
        self._data = collections.defaultdict(list)

        for list_type, messages in copy_data.items():
            try:
                # FIXME: remove this try..except block after we have forbidden
                #  to create Accumulatable class without LIST_CLASS.
                await self._sink.process_message(list_type(items=messages))
            except TypeError:
                logger.error("%s, %s", list_type, messages)
                raise
