import asyncio
import concurrent.futures
import contextlib
import logging
import os
import uuid
from typing import Generator, List

from defence360agent.api.server import APIError, send_message
from defence360agent.contracts import license
from defence360agent.contracts.config import Core
from defence360agent.contracts.messages import Message, Reportable
from defence360agent.contracts.plugins import MessageSink, expect
from defence360agent.utils import RecurringCheckStop, recurring_check

logger = logging.getLogger(__name__)


class SendToServer(MessageSink):
    """Send messages to server.

    * process Reportable messages;
    * add them to a pending messages list;
    * send all pending messages to server when list is full (contains
      _PENDING_MESSAGES_LIMIT items or more);
    * send all pending messages on plugin shutdown."""
    # TODO: consider reducing the limit,
    # since it can lead to messages loss on shutdown (DEF-20261)
    _PENDING_MESSAGES_LIMIT = int(
        os.environ.get("IMUNIFYAV_MESSAGES_COUNT_TO_SEND", 50)
    )
    _SHUTDOWN_SEND_TIMEOUT = 60  # 1 minute

    async def create_sink(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self._pending = []  # type: List[Message]
        self._try_send = asyncio.Event()
        self._stopping = asyncio.Event()
        self._sender_task = loop.create_task(self._sender())

    async def shutdown(self) -> None:
        self._stopping.set()
        self._try_send.set()
        try:
            await asyncio.wait_for(self._sender_task,
                                   self._SHUTDOWN_SEND_TIMEOUT)
        except asyncio.TimeoutError:
            logger.warning(
                "Timeout (%ds) sending messages to server on shutdown.",
                self._SHUTDOWN_SEND_TIMEOUT,
            )
            with contextlib.suppress(asyncio.CancelledError):
                await self._sender_task
        if self._pending:
            logger.warning('Dropped %s messages not sent to server',
                           len(self._pending))
            logger.warning("Dropped queue %r", self._pending)

    @contextlib.contextmanager
    def _get_api(self) -> Generator[send_message.SendMessageAPI, None, None]:
        base_url = os.environ.get('IMUNIFYAV_API_BASE')
        # we send messages sequentially, so max_workers=1
        with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
            api = send_message.SendMessageAPI(Core.VERSION, base_url,
                                              executor=executor)
            api.set_product_name(license.LicenseCLN.get_product_name())
            api.set_server_id(license.LicenseCLN.get_server_id())
            api.set_license(license.LicenseCLN.get_token())
            yield api

    @expect(Reportable)
    async def send_to_server(self, message: Message) -> None:
        if 'message_id' not in message:
            message['message_id'] = uuid.uuid4().hex
        self._pending.append(message)
        self._try_send.set()

    @recurring_check(0)
    async def _sender(self) -> None:
        await self._try_send.wait()
        self._try_send.clear()

        if (self._stopping.is_set()
                or len(self._pending) >= self._PENDING_MESSAGES_LIMIT):
            await self._send_pending_messages()

        if self._stopping.is_set():
            # there can still be pending messages but we don't care.
            # we care about common case where messages come infrequently
            # and we should send them on shutdown.
            raise RecurringCheckStop()

    async def _send_pending_messages(self) -> None:
        messages, self._pending = self._pending, []
        unsuccessful = []
        logger.info("Sending %s messages", len(messages))
        with self._get_api() as api:
            if api.server_id is not None:
                for message in messages:
                    try:
                        await api.send_message(message)
                        logger.info("message sent %s", {
                            "method": message.get("method"),
                            "message_id": message.get("message_id"),
                        })
                    except APIError as exc:
                        logger.warning(
                            'Failed to send message %s to server: %s',
                            message,
                            exc
                        )
                        unsuccessful.append(message)
        if unsuccessful:
            logger.info("Unsuccessful to send %s messages", len(unsuccessful))
        self._pending = unsuccessful + self._pending
