from defence360agent.contracts.hook_events import HookEvent
from defence360agent.contracts.messages import MalwareCleanup, MalwareScan
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.hooks.execute import execute_hooks
from defence360agent.malwarelib.config import MalwareHitStatus
from defence360agent.malwarelib.model import MalwareHit

EVENTS = (
    HookEvent.AgentStarted,
    HookEvent.AgentMisconfig,
    HookEvent.LicenseExpired,
    HookEvent.LicenseExpiring,
    HookEvent.LicenseRenewed,
    HookEvent.MalwareScanningStarted,
    HookEvent.MalwareScanningFinished,
    HookEvent.MalwareCleanupStarted,
    HookEvent.MalwareCleanupFinished,
    HookEvent.MalwareDetectedCritical,
)


async def detected_hook(
    sink, scan_id, scan_type, started, path, total, error=None
):
    total_malicious, hits = MalwareHit.malicious_list(
        by_scan_id=scan_id, ignore_cleaned=True
    )
    if total_malicious:
        malware_detected = HookEvent.MalwareDetectedCritical(
            scan_id=scan_id,
            scan_type=scan_type,
            error=error,
            started=started,
            path=path,
            total_files=total,
            total_malicious=total_malicious,
            DUMP=hits,
        )
        await sink.process_message(malware_detected)


class EventHookExecutor(MessageSink, MessageSource):
    PROCESSING_ORDER = MessageSink.ProcessingOrder.EVENT_HOOK

    async def create_sink(self, loop):
        self._loop = loop

    async def create_source(self, loop, sink):
        self._loop = loop
        self._sink = sink

    @expect(*EVENTS)
    async def receive_event(self, event):
        self._loop.create_task(execute_hooks(event))

    @expect(MalwareCleanup)
    async def process_malware_cleanup(self, message):
        error = message['error']
        status = 'failed' if error else 'ok'

        hits = message['hits']
        cleaned = [h for h in hits if h.status in MalwareHitStatus.RESTORABLE]
        dump = [hit.as_dict() for hit in hits]

        malware_cleanup = HookEvent.MalwareCleanupFinished(
            cleanup_id=message['cleanup_id'],
            started=message['started'],
            error=error,
            total_files=len(hits),
            total_cleaned=len(cleaned),
            status=status,
            DUMP=dump,
        )
        await self._sink.process_message(malware_cleanup)

    @expect(MalwareScan)
    async def process_malware_scan(self, message):
        summary = message['summary']

        # Message without summary means that detached scan is finished
        #  and summary will arrive along with results in another message - this
        #  is when the hooks should be called.
        if not summary.get('path'):
            return

        await detected_hook(
            self._sink,
            summary['scanid'],
            summary["type"],
            summary['started'],
            summary['path'],
            summary['total_files'],
            summary['error'],
        )
