"""
Generic Sensor plugin

- Creates listening unix domain socket on config.GenericSensor.SOCKET_PATH
- Expects alert data formatted as

"""
import asyncio
import base64
import json
import os
import time
from logging import getLogger

from defence360agent.contracts.config import SimpleRpc
from defence360agent.contracts.messages import (
    MSGS_WITHOUT_IP,
    MalwareCleanComplete,
    MalwareRestoreComplete,
    MalwareScan,
    MalwareScanComplete,
    MalwareScanTask,
    Noop,
)
from defence360agent.contracts.plugins import Sensor
from defence360agent.internals.global_scope import g
from defence360agent.internals.logger import getNetworkLogger
from defence360agent.simple_rpc import RpcServerAV
from defence360agent.utils.buffer import LineBuffer

logger, network_logger = getLogger(__name__), getNetworkLogger(__name__)


class Protocol(asyncio.Protocol):
    METHOD2MSGTYPE = {
        "NOOP": Noop,
        "MALWARE_SCAN": MalwareScan,
        "MALWARE_SCAN_TASK": MalwareScanTask,
        "MALWARE_SCAN_COMPLETE": MalwareScanComplete,
        "MALWARE_CLEAN_COMPLETE": MalwareCleanComplete,
        "MALWARE_RESTORE_COMPLETE": MalwareRestoreComplete,
    }

    def __init__(self, loop, sink, *_):
        self._loop = loop
        self._sink = sink
        self._line_buffer = LineBuffer()

    def connection_made(self, transport):
        network_logger.debug("Connection made")

    def data_received(self, data):
        msgs = data.decode()
        if not msgs.strip():
            logger.error('Empty message received <%s>', msgs)
            return

        self._line_buffer.append(msgs)
        for msg in self._line_buffer:
            if msg:
                network_logger.debug("data_received: {!r}".format(msg))
                tokens = self._parse_msg(msg)
                if tokens:
                    tokens['timestamp'] = time.time()
                    self._process_msg(tokens)

    def _parse_msg(self, msg):
        try:
            return json.loads(msg)
        except json.JSONDecodeError:
            logger.exception("data_received(%s): unable to decode", repr(msg))

    def _process_msg(self, tokens):
        # map 'method' to appropriate Message type
        try:
            msgtype = self.METHOD2MSGTYPE[tokens['method']]
        except KeyError as e:
            logger.error("data_received(%s): Wrong or missing 'method' [%s]",
                         repr(tokens), repr(e))
            return

        if tokens['method'] == 'MALWARE_SCAN_TASK':
            tokens['filelist'] = [
                os.fsdecode(base64.b64decode(f)) for f in tokens['filelist']
            ]
        elif tokens['method'] == 'SYNCLIST':  # added to avoid using pyrasite
            logger.info('Received test SynclistResponse')
        elif not tokens.get('attackers_ip') and \
                tokens['method'] not in MSGS_WITHOUT_IP:
            logger.error('Method type is %s but empty or no <attackers_ip> '
                         'in message <%s>', tokens.get('method'), tokens)
            return

        self._loop.create_task(
            self._sink.process_message(msgtype(tokens))
        )

    def connection_lost(self, transport):
        network_logger.debug("Disconnected")


class GenericSensor(Sensor):
    SOCKET_PATH = '/var/run/defence360agent/generic_sensor.sock.2'
    PROTOCOL_CLASS = Protocol

    async def create_sensor(self, loop, sink):
        if SimpleRpc.SOCKET_ACTIVATION:
            class GenericSensorSocket(RpcServerAV):
                SOCKET_PATH = self.SOCKET_PATH
                PROTOCOL_CLASS = self.PROTOCOL_CLASS

            g.sensor_server = await GenericSensorSocket.create(loop, sink)
            return g.sensor_server
        else:
            # FIXME make sure root can write to
            os.makedirs(os.path.dirname(self.SOCKET_PATH), exist_ok=True)
            if os.path.exists(self.SOCKET_PATH):
                os.unlink(self.SOCKET_PATH)

            g.sensor_server = await loop.create_unix_server(
                lambda: self.PROTOCOL_CLASS(loop, sink),
                self.SOCKET_PATH)
            return g.sensor_server
