import asyncio
import logging
import time
from collections import Counter
from pathlib import Path
from typing import List

from defence360agent.api.server import APIError
from defence360agent.api.server.events import EventsAPI
from defence360agent.contracts.config import (
    Core,
    IContactMessageType,
    MalwareScanSchedule
)
from defence360agent.contracts.hook_events import HookEvent
from defence360agent.contracts.messages import IContactSent
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.internals.the_sink import TheSink
from defence360agent.malwarelib.config import MalwareScanType
from defence360agent.model.icontact import IContactThrottle, THROTTLING_PERIOD
from defence360agent.subsys.panels.cpanel import cPanel
from defence360agent.subsys.panels.hosting_panel import HostingPanel
from defence360agent.utils import (
    await_for,
    create_task_and_log_exceptions,
    recurring_check,
    retry_on,
)
from defence360agent.utils.common import DAY

logger = logging.getLogger(__name__)


class IContactSender(MessageSink, MessageSource):
    PROCESSING_ORDER = MessageSink.ProcessingOrder.ICONTACT_SENT

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tasks = []
        self._notification_flag_path = (
            Path(Core.TMPDIR) / "icontact_generic_notifications"
        )

    async def create_sink(self, loop):
        pass

    async def _send_icontact_message(
        self, *, message_type, params, period_limit
    ):
        if message_type is None:
            return
        if not IContactThrottle.may_be_notified(message_type, period_limit):
            return
        template_args = await self._panel.notify(
            message_type=next(
                (
                    type_ for type_ in IContactMessageType
                    if type_ == message_type
                ),
                IContactMessageType.GENERIC
            ), params=params
        )
        if template_args:
            IContactThrottle.refresh(message_type)
            sent_message = IContactSent(
                message_type=message_type,
                timestamp=int(time.time()),
                template_args=template_args,
            )
            await self._sink.process_message(sent_message)

    async def create_source(self, loop, sink: TheSink):
        self._sink = sink
        self._panel = HostingPanel()
        if self._panel.NAME == cPanel.NAME:
            self._tasks = [
                create_task_and_log_exceptions(loop, coro)
                for coro in [self.no_scan_schedule, self.generic_notifications]
            ]

    async def shutdown(self):
        for task in self._tasks:
            task.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)

    @expect(HookEvent.MalwareDetectedCritical)
    async def malware_found(self, event):
        if event["scan_type"] != MalwareScanType.BACKGROUND:
            return

        if self._panel.NAME != cPanel.NAME:
            return

        hits: List[dict] = event.get("DUMP")
        if not hits:
            return

        counter = Counter(hit["username"] for hit in hits)
        infected_data = counter.most_common(1)
        infected_user = infected_data[0][0]

        await self._send_icontact_message(
            message_type=IContactMessageType.MALWARE_FOUND,
            params={"infected_user": infected_user},
            period_limit=THROTTLING_PERIOD[
                IContactMessageType.MALWARE_FOUND
            ],
        )

    @recurring_check(DAY)
    async def no_scan_schedule(self):
        try:
            cron_contains_scheduled_scan = (
                MalwareScanSchedule.CMD in
                Path(MalwareScanSchedule.CRON_PATH).read_text()
            )
        except FileNotFoundError:
            cron_contains_scheduled_scan = False

        if not cron_contains_scheduled_scan:
            await self._send_icontact_message(
                message_type=IContactMessageType.SCAN_NOT_SCHEDULED,
                params={},
                period_limit=THROTTLING_PERIOD[
                    IContactMessageType.SCAN_NOT_SCHEDULED
                ],
            )

    @retry_on(APIError, on_error=await_for(seconds=10), max_tries=3)
    async def get_notifications(self) -> list:
        notifications = []
        if (
            not self._notification_flag_path.exists() or
            (self._notification_flag_path.stat().st_mtime + DAY) < time.time()
        ):  # send notification request no more than once a day
            notifications = await EventsAPI.notification()
            # update flag modify time
            self._notification_flag_path.touch(mode=0o644, exist_ok=True)
        return notifications

    @recurring_check(DAY)
    async def generic_notifications(self):
        if (notifications := await self.get_notifications()):
            logger.info(
                "Sending %s generic icontact notifications", len(notifications)
            )
            for notification in notifications:
                await self._send_icontact_message(
                    message_type=notification["type"],
                    params={
                        "subject": notification["notification_subject"],
                        "body_html": notification["notification_body_html"],
                    },
                    period_limit=notification["notification_period_limit"],
                )
