"""Gather information from Plesk via DB querries.
"""

import logging
from collections import defaultdict
from typing import List, Dict, Sequence

from defence360agent.utils import retry_on, check_run, CheckRunError

logger = logging.getLogger(__name__)


@retry_on(CheckRunError, max_tries=3)
async def _run_query(query):
    return (await check_run(['plesk', 'db', '-N', '-e', query])).decode()


async def get_user_to_domain() -> Dict[str, List[str]]:
    """Return mapping: user -> user's domains. """

    result = (await _run_query(
        "select login, name from domains "
        "   left join hosting on dom_id = domains.id "
        "   right join sys_users on hosting.sys_user_id = sys_users.id"
    )).split()
    result_mapping = defaultdict(list)
    for user, domain in zip(result[0::2], result[1::2]):
        if domain != 'NULL':
            result_mapping[user].append(domain)
    return result_mapping


async def get_domain_to_user() -> Dict[str, List[str]]:
    """Return mapping: domain -> user. """

    result = (await _run_query(
        "select name, login "
        "from domains "
        "   left join hosting on dom_id = domains.id "
        "   left join sys_users on hosting.sys_user_id = sys_users.id"
    )).split()
    return {
        domain: [user] for domain, user in zip(result[0::2], result[1::2])
    }


async def get_domains() -> List[str]:
    """Return: list of domains"""

    return (await _run_query(
        "select name from domains"
    )).split()


async def get_users() -> List[str]:
    """Return: users that created by Plesk"""

    return (await _run_query(
        "SELECT sys_users.login FROM sys_users "
        "JOIN hosting ON hosting.sys_user_id=sys_users.id "
        "JOIN domains ON hosting.dom_id=domains.id AND domains.webspace_id=0"
    )).split()


async def count_customers_with_subscriptions() -> int:
    """Return: count of customers with subscriptions

        If customer (admin or reseller) do not have subscription,
        we do not count him. Type of subscription dosen't matter.
        We count customer only once, regardless of quantity of
        subscriptions.

    """

    return int(await _run_query("select count(distinct cl_id) from domains"))


async def get_admin_emails() -> list:
    emails = await _run_query("SELECT email FROM clients WHERE type='admin';")
    return [email for email in emails.split('\n') if email]


async def list_docroots() -> Sequence[str]:
    query = "SELECT DISTINCT hosting.www_root FROM hosting;"
    return (await _run_query(query)).split()


async def list_docroots_domains_users():
    sql = "select hosting.www_root, domains.name, sys_users.login" \
          " from hosting" \
          " inner join domains on hosting.dom_id = domains.id" \
          " inner join sys_users on hosting.sys_user_id=sys_users.id"
    data = await _run_query(sql)
    retval = [string.split() for string in data.split('\n') if string]
    return retval
