import abc
from collections import namedtuple
from typing import Dict, FrozenSet, Iterable, List, Optional, Set

from im360.utils.validate import IP, IPVersion

from .libipset import IPSetRestoreCmd

IP_SET_PREFIX = "i360"


IPSetCount = namedtuple("IPSetCount", ["name", "db_count", "ipset_count"])


def get_ipset_family(ip_version: IPVersion):
    assert ip_version in (
        IP.V4,
        IP.V6,
    ), f"ip version {ip_version} is incorrect"
    return "inet6" if ip_version == IP.V6 else "inet"


class AbstractIPSet(abc.ABC):
    """Entity to manage a specific slice of iptables rules & ipsets.

    See ..RuleSet
    """

    @abc.abstractmethod
    async def get_all_ipsets(self, ip_version: IPVersion) -> FrozenSet[str]:
        pass

    @abc.abstractmethod
    async def get_rules(
        self, ip_version: IPVersion, **kwargs
    ) -> Iterable[dict]:
        pass

    @abc.abstractmethod
    async def restore(self, ip_version: IPVersion) -> None:
        pass

    @abc.abstractmethod
    async def gen_ipset_create_ops(self, ip_version: IPVersion) -> List[str]:
        pass

    async def gen_ipset_destroy_ops(
        self, ip_version: IPVersion, existing: Set[str]
    ) -> Optional[Dict[str, IPSetRestoreCmd]]:
        """Generate specific destroy `ipset restore` commands.

        Return None if no preference.
        """
        return None

    @abc.abstractmethod
    async def get_ipsets_count(
        self, ip_version: IPVersion
    ) -> List[IPSetCount]:  # pragma: no cover
        pass
