import asyncio
import logging
from typing import Dict, FrozenSet, Iterable, List

from defence360agent.contracts.config import PORT_BLOCKING_MODE_ALLOW
from defence360agent.model.simplification import run_in_executor
from defence360agent.utils import retry_on, timeit
from im360.contracts.config import Firewall
from im360.internals.core import ip_versions
from im360.internals.core.firewall import FirewallRules, get_firewall
from im360.internals.core.firewall.base import FirewallBatchCommandError
from im360.internals.core.ipset import (
    IP_SET_PREFIX,
    AbstractIPSet,
    IPSetCount,
    get_ipset_family,
    libipset,
)
from im360.internals.core.ipset.libipset import IPSetCmdBuilder
from im360.model.firewall import BlockedPort, IgnoredByPort
from im360.model.firewall import IPv4 as DB_IPv4
from im360.model.firewall import IPv6 as DB_IPv6
from im360.utils.validate import IP, IPVersion

from .redirect import IPSetNoRedirectPort

logger = logging.getLogger(__name__)


class IPSetPort(AbstractIPSet):
    TEMPLATE = "{prefix}.{ip_version}.ignored-by-{proto}-{port}"

    TCP = "tcp"
    UDP = "udp"
    ALL = "all"

    PROTOS = (TCP, UDP, ALL)
    MIN_PORT, MAX_PORT = 0, 65535

    def expand_proto(self, proto):
        assert proto in self.PROTOS, 'protocol "{}" is not supported'.format(
            proto
        )

        if proto == self.ALL:
            return [self.TCP, self.UDP]
        return [proto]

    async def block(self, item, *args, **kwargs):
        port, generic_proto = item

        for ip_version in ip_versions.enabled():
            async with await get_firewall(ip_version) as fw:
                set_name = self.generate_set_name(
                    ip_version, port, generic_proto
                )

                await libipset.create_hash_set(
                    set_name,
                    timeout=0,
                    set_type=libipset.HASH_NET,
                    family=get_ipset_family(ip_version),
                )

                actions = [
                    fw.append_rule(
                        FirewallRules.port_rule(
                            set_name,
                            port,
                            proto,
                            policy=FirewallRules.LOG_BLOCK_PORT_CHAIN,
                        ),
                        chain=FirewallRules.BP_INPUT_CHAIN,
                        priority=FirewallRules.DEFAULT_PRIORITY,
                    )
                    for proto in self.expand_proto(generic_proto)
                ]

                decorated = retry_on(FirewallBatchCommandError, max_tries=3)(
                    fw.commit
                )
                await decorated(actions)
                await IPSetNoRedirectPort().add_item(port, ip_version)

    async def unblock(self, item, *args, **kwargs):
        port, generic_proto = item

        for ip_version in ip_versions.enabled():
            async with await get_firewall(ip_version) as fw:
                set_name = self.generate_set_name(
                    ip_version, port, generic_proto
                )
                actions = [
                    fw.delete_rule(
                        FirewallRules.port_rule(
                            set_name,
                            port,
                            proto,
                            policy=FirewallRules.LOG_BLOCK_PORT_CHAIN,
                        ),
                        chain=FirewallRules.BP_INPUT_CHAIN,
                        priority=FirewallRules.DEFAULT_PRIORITY,
                    )
                    for proto in self.expand_proto(generic_proto)
                ]
                decorated = retry_on(FirewallBatchCommandError, max_tries=3)(
                    fw.commit
                )
                await decorated(actions)
                await IPSetNoRedirectPort().delete_item(port, ip_version)

            set_name = self.generate_set_name(ip_version, port, generic_proto)
            await libipset.delete_set(set_name)

    async def gen_restore_ops(self, ip_version: str):
        return []

    async def gen_ipset_create_ops(self, ip_version: IPVersion) -> List[str]:
        """
        Generate list of commands to create all ip sets

        :return: list of ipset commands to use with ipset restore
        """
        result: List[str] = []
        if Firewall.port_blocking_mode != PORT_BLOCKING_MODE_ALLOW:
            return result
        for item in await self._fetch():
            port, generic_proto = item
            set_name = self.generate_set_name(ip_version, port, generic_proto)
            result.append(
                IPSetCmdBuilder.get_create_cmd(
                    set_name, family=get_ipset_family(ip_version)
                )
            )
        return result

    async def _fetch(self):
        return await run_in_executor(
            asyncio.get_event_loop(),
            lambda: [(row.port, row.proto) for row in BlockedPort.select()],
        )

    @classmethod
    def generate_set_name(cls, ip_version: str, port: int, proto: str) -> str:
        return cls.TEMPLATE.format(
            prefix=IP_SET_PREFIX, ip_version=ip_version, proto=proto, port=port
        )

    async def get_all_ipsets(self, ip_version: str) -> FrozenSet[str]:
        if Firewall.port_blocking_mode != PORT_BLOCKING_MODE_ALLOW:
            return frozenset()
        return frozenset(
            self.generate_set_name(ip_version, port, gen_proto)
            for port, gen_proto in await self._fetch()
        )

    async def get_rules(self, ip_version: str, **kwargs) -> Iterable[dict]:
        if Firewall.port_blocking_mode != PORT_BLOCKING_MODE_ALLOW:
            return []
        result = [
            dict(
                rule=FirewallRules.compose_action(
                    FirewallRules.BP_INPUT_CHAIN
                ),
                chain=FirewallRules.IMUNIFY_INPUT_CHAIN,
                table=FirewallRules.FILTER,
                priority=FirewallRules.PORT_PROTO_PRIORITY,
            )
        ]
        for item in await self._fetch():
            port, generic_proto = item
            set_name = self.generate_set_name(ip_version, port, generic_proto)
            result.extend(
                [
                    dict(
                        rule=FirewallRules.port_rule(
                            set_name,
                            port,
                            proto,
                            policy=FirewallRules.LOG_BLOCK_PORT_CHAIN,
                        ),
                        chain=FirewallRules.BP_INPUT_CHAIN,
                        table=FirewallRules.FILTER,
                        priority=FirewallRules.DEFAULT_PRIORITY,
                    )
                    for proto in self.expand_proto(generic_proto)
                ]
            )
        return result

    async def restore(self, ip_version: str) -> None:
        if Firewall.port_blocking_mode != PORT_BLOCKING_MODE_ALLOW:
            return
        with timeit("ipset_restore", logger):
            await libipset.restore(await self.gen_restore_ops(ip_version))

    async def get_ipsets_count(self, ip_version: str) -> list:
        return []


class IPSetIgnoredByPort(AbstractIPSet):
    async def block(self, ip, port, proto, *args, **kwargs):
        set_name = IPSetPort.generate_set_name(IP.type_of(ip), port, proto)
        await libipset.add_item(set_name, ip, timeout=0)

    async def unblock(self, ip, port, proto, *args, **kwargs):
        set_name = IPSetPort.generate_set_name(IP.type_of(ip), port, proto)
        await libipset.delete_item(set_name, ip)

    async def gen_restore_ops(self, ip_version: str):
        """
        Generate list of commands to fill all ip sets

        :return: list of ipset commands to use with ipset restore
        """
        add_template = "add {set_name} {ip_net} -exist"

        lines = []
        for ip, port, proto in await self._fetch():
            version = IP.type_of(ip)
            if version == ip_version:
                lines.append(
                    add_template.format(
                        set_name=IPSetPort.generate_set_name(
                            version, port, proto
                        ),
                        ip_net=ip,
                    )
                )
        return lines

    async def _fetch(self, version=None):
        return await run_in_executor(
            asyncio.get_event_loop(),
            lambda: [
                (row.ip, row.port_proto.port, row.port_proto.proto)
                for row in IgnoredByPort.fetch(version)
            ],
        )

    async def gen_ipset_create_ops(self, ip_version: IPVersion) -> List[str]:
        # Just to follow an interface, there are no ipsets for this entity
        return []

    async def get_all_ipsets(self, ip_version: str) -> FrozenSet[str]:
        return frozenset()

    async def get_rules(self, ip_version: str, **kwargs) -> Iterable[dict]:
        return []

    def is_enabled(self):
        return Firewall.port_blocking_mode == PORT_BLOCKING_MODE_ALLOW

    async def restore(self, ip_version: str) -> None:
        if not self.is_enabled():
            return

        with timeit("ipset_restore", logger):
            await libipset.restore(await self.gen_restore_ops(ip_version))

    async def get_ipsets_count(self, ip_version: str) -> list:
        ipsets = []
        if self.is_enabled():
            ipsets_count: Dict[str, int] = {}  # ipset name -> ips count
            db_ip_version = DB_IPv4 if ip_version == IP.V4 else DB_IPv6
            for ip, port, proto in await self._fetch(db_ip_version):
                set_name = IPSetPort.generate_set_name(
                    ip_version, port, proto
                )
                ipsets_count[set_name] = ipsets_count.setdefault(
                    set_name, 0) + 1

            for set_name, expected_count in ipsets_count.items():
                ipset_count = await libipset.get_ipset_count(set_name)
                ipsets.append(
                    IPSetCount(
                        name=set_name,
                        db_count=expected_count,
                        ipset_count=ipset_count
                    )
                )
        return ipsets
