import logging
from abc import abstractmethod
from enum import Enum
from typing import FrozenSet, Iterable, List

from defence360agent.contracts.config import PORT_BLOCKING_MODE_DENY
from im360.contracts.config import Firewall
from im360.internals.core.firewall import FirewallRules
from im360.internals.core.ipset import (
    IP_SET_PREFIX,
    AbstractIPSet,
    IPSetCount,
    libipset,
)
from im360.internals.strategy import Strategy
from im360.utils.net import IP, TCP, UDP
from im360.utils.validate import IPVersion

logger = logging.getLogger(__name__)


class TrafficDirection(Enum):
    INPUT = 'input'
    OUTPUT = 'output'


WEBSHIELD_PORTS = ['52223', '52224', '52233', '52234']  # tcp in
ACRONIS_PORTS = ['44445', '55556', '7770-7800']  # tcp out


class PortBlockingDenyModeIPSetManager:
    TEMPLATE = '{prefix}.{ip_version}.{dir}-ports-{proto}'

    def __init__(self, direction, proto, ip_tables_manager):
        self.direction = direction
        self.proto = proto
        self.ip_tables_manager = ip_tables_manager

    async def restore(self, ip_version: str) -> None:
        name = self.get_set_name(ip_version)
        await libipset.flush_set(name)
        lines = []
        for port in self.fetch(ip_version):
            lines.append(
                ' '.join(libipset.prepare_ipset_command('add', name, port)),
            )
        await libipset.restore(name=name, lines=lines)

    async def gen_ipset_create_ops(self, ip_version: IPVersion) -> str:
        return 'create {name} bitmap:port range 0-65535 timeout 0 -exist'\
            .format(name=self.get_set_name(ip_version))

    def fetch(self, ip_version):
        return self.ip_tables_manager.get_config_option(ip_version,
                                                        self.proto)

    def get_set_name(self, ip_version):
        return self.TEMPLATE.format(prefix=IP_SET_PREFIX,
                                    dir=self.direction,
                                    proto=self.proto,
                                    ip_version=ip_version)


class PortBlockingDenyModeIPSet(AbstractIPSet):
    def __init__(self, direction):
        self.direction = direction
        self.ip_sets = [
            PortBlockingDenyModeIPSetManager(direction, proto, self)
            for proto in (TCP, UDP)
        ]

    @abstractmethod
    def get_config_option(self, ip_version, proto):
        raise NotImplementedError

    @abstractmethod
    def get_chain_name(self):
        raise NotImplementedError

    @abstractmethod
    def get_loopback_rule(self):
        raise NotImplementedError

    @abstractmethod
    def get_parent_chain_name(self):
        raise NotImplementedError

    @abstractmethod
    def get_block_action(self):
        raise NotImplementedError

    async def get_rules(self, ip_version: str, **kwargs) -> Iterable[dict]:
        if not self._enabled(ip_version):
            return []
        chain_name = self.get_chain_name()
        rules = [
            self.get_loopback_rule(),
            ('-p', 'icmp', '-j', 'RETURN'),
            ('-m', 'conntrack', '--ctstate', 'RELATED,ESTABLISHED',
             '-j', 'RETURN'),
            *[('-p', ip_set.proto, '-m', 'set', '--match-set',
               ip_set.get_set_name(ip_version), 'dst', '-j', 'RETURN')
              for ip_set in self.ip_sets
              ],
            self.get_block_action()
        ]
        return [
            *[
                dict(
                    rule=rule,
                    chain=chain_name,
                    priority=priority,
                ) for priority, rule in enumerate(rules)
            ],
            dict(
                rule=('-j', chain_name),
                chain=self.get_parent_chain_name(),
                priority=FirewallRules.PORT_PROTO_PRIORITY,
            ),
        ]

    def _enabled(self, ip_version):
        return (
            Firewall.port_blocking_mode == PORT_BLOCKING_MODE_DENY
            and ip_version != IP.V6
            and Strategy.current != Strategy.CSF_COOP_STRATEGY
        )

    async def restore(self, ip_version: str) -> None:
        if self._enabled(ip_version):
            for ip_set in self.ip_sets:
                await ip_set.restore(ip_version)

    async def gen_ipset_create_ops(self, ip_version: IPVersion) -> List[str]:
        result = []
        if self._enabled(ip_version):
            for ip_set in self.ip_sets:
                result.append(await ip_set.gen_ipset_create_ops(ip_version))
        return result

    async def get_all_ipsets(self, ip_version: str) -> FrozenSet[str]:
        result = []
        if self._enabled(ip_version):
            for ip_set in self.ip_sets:
                result.append(ip_set.get_set_name(ip_version))
        return frozenset(result)

    async def get_ipsets_count(self, ip_version: str) -> list:
        ipsets = []
        if self._enabled(ip_version):
            for ip_set in self.ip_sets:
                set_name = ip_set.get_set_name(ip_version)
                expected_count = len(ip_set.fetch(ip_version))
                ipset_count = await libipset.get_ipset_count(set_name)
                ipsets.append(
                    IPSetCount(
                        name=set_name,
                        db_count=expected_count,
                        ipset_count=ipset_count
                    )
                )
        return ipsets


class InputPortBlockingDenyModeIPSet(
    PortBlockingDenyModeIPSet,
):
    def get_block_action(self):
        return FirewallRules.compose_action(FirewallRules.LOG_BLOCK_PORT_CHAIN)

    def __init__(self):
        super().__init__(TrafficDirection.INPUT.value)

    def get_config_option(self, ip_version, proto):
        if ip_version == IP.V6:
            return []
        if proto == TCP:
            return Firewall.TCP_IN_IPV4 + WEBSHIELD_PORTS
        if proto == UDP:
            return Firewall.UDP_IN_IPV4
        raise NotImplementedError()

    def get_chain_name(self):
        return FirewallRules.BP_INPUT_CHAIN

    def get_parent_chain_name(self):
        return FirewallRules.IMUNIFY_INPUT_CHAIN

    def get_loopback_rule(self):
        return ('-i', 'lo', '-j', 'RETURN')


class OutputPortBlockingDenyModeIPSet(
    PortBlockingDenyModeIPSet,
):
    def get_block_action(self):
        return FirewallRules.compose_action(
            FirewallRules.REJECT
        )

    def __init__(self):
        super().__init__(TrafficDirection.OUTPUT.value)

    def get_config_option(self, ip_version, proto):
        if ip_version == IP.V6:
            return []
        if proto == TCP:
            return Firewall.TCP_OUT_IPV4 + ACRONIS_PORTS
        if proto == UDP:
            return Firewall.UDP_OUT_IPV4
        raise NotImplementedError()

    def get_chain_name(self):
        return FirewallRules.BP_OUTPUT_CHAIN

    def get_parent_chain_name(self):
        return FirewallRules.IMUNIFY_OUTPUT_CHAIN

    def get_loopback_rule(self):
        return ('-o', 'lo', '-j', 'RETURN')
