import asyncio
from logging import getLogger

from defence360agent import utils
from defence360agent.contracts.messages import ConfigUpdate
from defence360agent.contracts.plugins import MessageSink, expect
from im360.subsys.panels.base import (
    ModsecNotInstalledVendors,
    ModsecImunifyVendorNotInstalled,
)
from im360.contracts.config import Modsec
from im360.subsys.panels.base import use_modsec_lock
from im360.subsys.panels.hosting_panel import HostingPanel
from im360.subsys.panels.update_hooks import \
    update_account_compromise_prevention_rule_state

logger = getLogger(__name__)


class ModsecRulesetChecker(MessageSink):

    def __init__(self):
        self._loop = None
        self._task = None

    async def create_sink(self, loop):
        self._loop = loop
        self._task = self._loop.create_task(self.reinstall_vendor_if_needed())

    async def shutdown(self):
        self._task.cancel()
        await self._task

    @use_modsec_lock
    async def reinstall_vendor_if_needed(self):
        panel = HostingPanel()
        token = panel.installing_settings_var.set(True)
        try:
            try:
                # We can't do anything if Panel doesn't support modsec
                # or generate specific panel error
                installed_im360_vendor = await panel.get_i360_vendor_name()
            except (ModsecNotInstalledVendors,
                    ModsecImunifyVendorNotInstalled):
                # this means that we can't find any modsec vendors or
                # there is no im360 vendor
                installed_im360_vendor = None
            if installed_im360_vendor is None or \
                    Modsec.RULESET.lower() not in installed_im360_vendor:
                logger.info(
                    'Installed i360 vendor {} does not match expected type of '
                    'ruleset: {}\n Trying to reinstall modsec ruleset'.format(
                        str(installed_im360_vendor), str(Modsec.RULESET)))
                await panel.apply_modsec_files_update()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logger.error(
                'Something went wrong during reinstalling modsec ruleset: {}'.
                format(e))
        finally:
            panel.installing_settings_var.reset(token)

    @expect(ConfigUpdate)
    @utils.log_error_and_ignore()
    async def on_config_update(self, _):
        await self.reinstall_vendor_if_needed()

    @expect(ConfigUpdate)
    async def check_cms_account_compromise_prevention(self, _):
        """
        Update *WP_REDIRECT_CONF* to apply changes
        to *cms_account_compromise_prevention* config setting.
        """
        await update_account_compromise_prevention_rule_state()
