from typing import Dict

from functools import partial

from defence360agent.api import inactivity
from defence360agent.contracts.messages import RestoreFromBackupTask
from defence360agent.contracts.plugins import (
    expect,
    MessageSink,
    MessageSource,
)
from defence360agent.malwarelib.config import MalwareHitStatus
from defence360agent.malwarelib.model import MalwareHit
from im360.malwarelib.subsys.malware import MalwareAction

_group_by_status = partial(MalwareHit.group_by_attribute, attribute="status")


class RestoreFromBackupPlugin(MessageSink, MessageSource):
    """A plugin that handles RestoreFromBackupTask messages."""

    def __init__(self):
        self._loop, self._sink = None, None

    async def create_sink(self, loop):
        self._loop = loop

    async def create_source(self, loop, sink):
        self._sink = sink

    @expect(RestoreFromBackupTask)
    async def restore_from_backup(self, message: Dict):
        """Restores hits in RestoreFromBackupTask message from backup.

        - sets restore_from_backup_started status for received hits in
            the database.
        - successfully restored files are removed from the database.
        - statuses in the database are reverted from
            restore_from_backup_started to its original value for files that
            failed to restore.

        """
        hits = message.get("hits")
        if not hits:
            return
        initiator = message.get("initiator")

        MalwareHit.set_status(
            hits, MalwareHitStatus.RESTORE_FROM_BACKUP_STARTED
        )

        with inactivity.track.task("restore_from_backup"):
            to_restore = {}
            to_restore.setdefault(initiator, []).extend(hits)
            restored, failed = [], []

            for initiator, init_hits in to_restore.items():
                results = await MalwareAction.restore_from_backup(
                    hits=init_hits, initiator=initiator, sink=self._sink,
                )

                restored.extend(
                    [hit for hit, event in results.items() if event.successful]
                )

                failed.extend(
                    [
                        hit
                        for hit, event in results.items()
                        if not event.successful
                    ]
                )

            MalwareHit.delete_instances(restored)

            original_status = _group_by_status(failed)
            for status, failed_hits in original_status.items():
                MalwareHit.set_status(failed_hits, status)
