#!/opt/alt/python38/bin/python3
import time
from logging import getLogger

from defence360agent.contracts.plugins import LogStreamReader
from defence360agent.subsys.panels.base import PanelException
from defence360agent.utils import RecurringCheckStop, recurring_check
from im360.contracts.messages import SensorIncident
from im360.subsys import modsec_audit_log
from im360.subsys.panels.hosting_panel import HostingPanel

PLUGIN_ID = 'modsec'

logger = getLogger(__name__)


class ModsecSensor(LogStreamReader):
    async def create_sensor(self, loop, sink):
        self.source_file, audit_logdir = self._get_log_paths()
        if self.source_file is None:
            self._parser = None
        elif audit_logdir is None:
            self._parser = modsec_audit_log.SerialLogParser()
        else:
            self._parser = modsec_audit_log.RevolverParser(
                audit_logdir_path=audit_logdir)

        await super().create_sensor(loop, sink)

    def _get_log_paths(self):
        hp = HostingPanel()
        result = []
        for method in hp.get_audit_log_path, hp.get_audit_logdir_path:
            try:
                result.append(method())
            except PanelException as e:
                logger.warning("%s failed: %s", method, e)
                result.append(None)
        return tuple(result)

    @recurring_check(0)
    async def _infinite_read_and_proceed(self, stream_reader):
        try:
            bytes_ = await stream_reader.readline()

        # If limit was reached but no full line was consumed
        except ValueError as e:
            logger.warning(e)
            return

        if not bytes_:
            # eof
            raise RecurringCheckStop()

        try:
            tokens = self._parser.feed(bytes_)
            if not tokens:
                return
        except (modsec_audit_log.ParseError,
                modsec_audit_log.MalformedFileError) as e:
            logger.warning(*e.args)
            return

        ts = time.time()
        for incident in tokens:
            incident['timestamp'] = ts
            incident['domain'] = dict(
                incident.get('advanced', {}).get('headers', [])).get('Host')
            await self._sink.process_message(SensorIncident(incident))
