"""WebShield manager plugin.

It enables/disables webshield service on an imunify360 config change.

"""
import asyncio
import logging

from defence360agent.contracts import messages, plugins
from defence360agent import utils
from im360.contracts import config
from im360.subsys import webshield

__all__ = ["WebShieldManager"]

logger = logging.getLogger(__name__)


class WebShieldManager(plugins.MessageSink):

    async def create_sink(self, loop):
        self._lock = asyncio.Lock()
        await self._ensure_consistent_webshield_state()  # on start up

    @plugins.expect(messages.ConfigUpdate)
    async def on_config_update(
        self, message_ignored: messages.ConfigUpdate
    ):
        await self._ensure_consistent_webshield_state()

    @utils.log_error_and_ignore()
    @utils.retry_on(webshield.Error, max_tries=2)
    async def _ensure_consistent_webshield_state(self):
        async with self._lock:  # handle concurrent config updates
            should_be_running = config.Webshield.ENABLE
            if (
                await webshield.is_running()
            ) is not should_be_running:
                if should_be_running:
                    logger.info(
                        "WebShield is enabled in the config but it is not running."  # noqa: E501
                        " Enabling it..."
                    )
                    # enable on boot & start it right now
                    await webshield.service_enable(now=True)
                    logger.info("Enabled WebShield")
                else:
                    logger.info(
                        "WebShield is not enabled in the config but it is running."  # noqa: E501
                        " Disabling it..."
                    )
                    # disable on boot & stop it right now
                    await webshield.service_disable(now=True)
                    logger.info("Disabled WebShield")

            if should_be_running:
                await webshield.splashscreen_set_state(
                    config.Webshield.SPLASH_SCREEN)
