#!/opt/alt/python38/bin/python3
from defence360agent import server
from defence360agent.api import health
from defence360agent.contracts.config import Model
from defence360agent.contracts.license import LicenseCLN
from defence360agent.internals.cln import subscribe_to_license_changes
from defence360agent.internals.iaid import IndependentAgentIDAPI
from defence360agent.malwarelib.subsys.malware import (
    subscribe_to_malware_action,
)
from defence360agent.model import instance
from defence360agent.run import get_plugins as get_base_plugins
from im360 import rpc_handlers
from im360.application.settings import configure
from im360.internals.core import ip_versions
from im360.malwarelib.plugins import cleanup, detached_scan, store
from im360.malwarelib.subsys.malware import HackerTrapHitsSaver
from im360.plugins import (
    aggregate,
    aibolit_result_scan,
    backup_info_sender,
    cache_clear,
    cagefs,
    captcha_dos_detect,
    client360,
    config_set,
    conflicts,
    cpanel_uploader,
    db_auto_cleanup,
    fix_ip_address,
    graylist,
    group_ip_sync,
    ignored_rules,
    inotify,
    mod_sec_blocker,
    modsec_ruleset_checker,
    ossec_alert,
    pam,
    persistent_storage,
    php_immunity,
    post_action,
    remoteip_install,
    remote_iplist,
    restore_from_backup,
    restrict_graylisted,
    send_server_config,
    server_pull,
    serverpush_to_synclist,
    smtp_blocking,
    startup_actions,
    strategy_checker,
    ttl_graylist,
    update_files,
    waf_rules_configurator,
    webshield_captcha,
    webshield_manager,
    whitelist_current_user,
    whitelist_panels_login,
    malware_ignore_paths
)
from im360.plugins.protector import import_wblist, lazy_init, lfd, synclist
from im360.plugins.sensor import (
    cphulk,
    dos_detector,
    generic,
    ignore_alert_with_whitelisted_ip,
    modsec,
    ossec,
    search_bots,
)
from im360.plugins.sensor.webshield import WebshieldSensor
from im360.subsys import features


def get_plugins():
    return [
        # use lexicographical order (but don't rely on it in code)
        WebshieldSensor,
        aggregate.Aggregate,
        aibolit_result_scan.AibolitResultsScan,
        backup_info_sender.BackupInfoSender,
        cache_clear.CacheClear,
        cagefs.CageFS,
        captcha_dos_detect.CaptchaDosDetect,
        cleanup.CleanupDb,
        cleanup.RestoreOriginalDb,
        cleanup.ResultProcessor,
        client360.Client360,
        config_set.ConfigSet,
        conflicts.Conflicts,
        cpanel_uploader.CpanelUploadHookManager,
        cphulk.cpHulkSensor,
        db_auto_cleanup.DbCleanup,
        detached_scan.DetachedScanPlugin,
        dos_detector.DOSSensor,
        fix_ip_address.FixIPAddress,
        generic.GenericSensor,
        group_ip_sync.GroupIPSyncPlugin,
        graylist.ManageGrayList,
        ignore_alert_with_whitelisted_ip.IgnoreWhitelisted,
        ignore_alert_with_whitelisted_ip.IgnoreWhitelistedCSF,
        ignored_rules.FilterIgnoredRules,
        import_wblist.ExportWBList,
        inotify.InotifyScan,
        lazy_init.RealProtector,
        lfd.LFD,
        mod_sec_blocker.ModSecBlockBySeverity,
        modsec.ModsecSensor,
        modsec_ruleset_checker.ModsecRulesetChecker,
        ossec.JsonAlertReader,
        ossec_alert.OssecAlert,
        pam.PAM,
        webshield_captcha.WebshieldCaptcha,
        persistent_storage.PersistentStorage,
        post_action.PostAction,
        remoteip_install.RemoteIpInstall,
        remote_iplist.RemoteIPListPlugin,
        restore_from_backup.RestoreFromBackupPlugin,
        restrict_graylisted.SuppressGraylistedProcessing,
        search_bots.WhitelistSearchBots,
        send_server_config.SendServerConfig,
        server_pull.ServerPullFileSender,
        serverpush_to_synclist.ServerPushToSynclist,
        smtp_blocking.SMTPBlocker,
        startup_actions.StartupActions,
        store.StoreMalwareHits,
        strategy_checker.IDSSensor,
        synclist.SynclistPlugin,
        ttl_graylist.GraylistTimeout,
        update_files.UpdateFilesOnServerRequest,
        waf_rules_configurator.WAFRuleSetConfigurator,
        webshield_manager.WebShieldManager,
        whitelist_current_user.WhitelistCurrentUser,
        whitelist_panels_login.WhitelistPanelsLogin,
        php_immunity.PhpImmunityPlugin,
        malware_ignore_paths.MalwareIgnorePath,
    ]


async def update_health_sensor():
    if LicenseCLN.is_valid():
        health.sensor.registered()
    else:
        health.sensor.unregistered()


async def init_actions():
    # Any uncaught exceptions here prevents the agent from
    # starting. Non-critical (agent can continue) functionality should
    # either be moved to a plugin or caught&log *all* its exceptions.

    # Also nothing should block here indefinitely (any operation that
    # may block should have a timeout).

    ip_versions.init()
    subscribe_to_license_changes(features.update_repos)
    subscribe_to_license_changes(update_health_sensor)
    instance.db.execute_sql(
        "ATTACH '{}' AS proactive".format(Model.PROACTIVE_PATH))
    subscribe_to_malware_action("delete", HackerTrapHitsSaver.add_hit)
    subscribe_to_malware_action("cleanup", HackerTrapHitsSaver.add_hit)
    subscribe_to_license_changes(IndependentAgentIDAPI.activate)
    await IndependentAgentIDAPI.register()
    rpc_handlers.init()


def run():
    configure()
    plugins = get_plugins() + get_base_plugins()
    server.start(plugins, init_actions)


if __name__ == "__main__":
    run()
