from functools import wraps
from logging import getLogger

from defence360agent.contracts.config import ConfigFile, Core
from im360.contracts.config import Webshield
from im360.subsys.features.abstract_feature import (
    AbstractFeature,
    FeatureError,
)

logger = getLogger(__name__)


def require_webshield_enabled(fun):
    @wraps(fun)
    async def wrapper(*args, **kwargs):
        if not Webshield.ENABLE:
            raise FeatureError(
                "Webshield should be enabled to use this feature"
            )
        return await fun(*args, **kwargs)

    return wrapper


class InvisibleCaptcha(AbstractFeature):
    """
    This feature allows to activate/deactivate both aiohttp and Nginx captcha
    invisible option.
    """
    LOG_DIR = '/var/log/%s' % Core.PRODUCT
    INSTALL_LOG_FILE_MASK = '%s/dummy-install.log.*' % LOG_DIR
    REMOVE_LOG_FILE_MASK = '%s/dummy-remove.log.*' % LOG_DIR

    NAME = "invisible-captcha"

    def __init__(self, sink=None):
        super().__init__(sink)

    async def _check_installed_impl(self) -> bool:
        return Webshield.INVISIBLE_CAPTCHA

    @require_webshield_enabled
    async def check_installed(self) -> bool:
        return await self._check_installed_impl()

    @require_webshield_enabled
    @AbstractFeature.raise_if_shouldnt_install_now
    async def install(self):
        """
        :return path: to dummy log just to keep to the same usage
            convention
        """
        await self._set_invisible_captcha(True)
        return self.INSTALL_LOG_FILE_MASK

    @require_webshield_enabled
    @AbstractFeature.raise_if_shouldnt_remove_now
    async def remove(self):
        """
        :return path: to dummy log just to keep to the same usage
            convention
        """
        await self._set_invisible_captcha(False)
        return self.REMOVE_LOG_FILE_MASK

    async def _set_invisible_captcha(self, new_state):
        """
        :param bool new_state:

        :raise CalledProcessError:
        :raise OSError:
        """
        ConfigFile().set("WEBSHIELD", "invisible_captcha", new_state)
