import hashlib
import json
import logging
import os
from asyncio import CancelledError
from contextlib import suppress
from distutils.version import StrictVersion

from defence360agent.utils import (
    BACKUP_EXTENSION,
    CheckRunError,
    atomic_rewrite,
    file_hash,
)
from im360.files import MODSEC, Index
from im360.subsys import web_server
from im360.subsys.panels.base import APACHE
from im360.subsys.panels.generic.panel import GenericPanel

from .modsec_app_version_detector import map_components_versions_to_tags
from .panels.hosting_panel import HostingPanel
from .web_server import configtest, graceful_restart

logger = logging.getLogger(__name__)

COMPONENTS_VERSION_DB = (
    "/var/lib/cloudlinux-app-version-detector/components_versions.sqlite3"
)
RULES_CONF_PATTERN = "<IfModule security2_module>\n{}\n</IfModule>"
MAPPING_FILE = "tags_matching.json"


class NotSupportedWebserverError(Exception):
    pass


async def is_webserver_supported() -> bool:
    """Apache >= 2.4 is supported and security2_module installed.

    - litespeed is not supported
    - nginx is not supported

    Apache is expected to be running, otherwise False is returned
    In case of any error, False is returned also
    """
    hp = HostingPanel()
    webserver = hp.get_web_server()
    if isinstance(hp, GenericPanel):
        # on Generic panel we expect that apache version >= 2.4 will be used
        # no check apache running and version explicitly
        return webserver == APACHE

    if webserver != APACHE:
        return False
    try:
        version = await web_server.apache_version()
        modules = await web_server.apache_modules()
    except CancelledError:
        raise
    except Exception as exc:
        logger.error("Error occurs while getting Apache version: %s", exc)
        return False
    return version >= StrictVersion("2.4") and b'security2_module' in modules


async def update_waf_rules_config():
    """
    Update modsec config file with enabled tags for specific directories.
    """
    if not await is_webserver_supported():
        raise NotSupportedWebserverError(
            "WAF rules configurator supports only apache webserver with "
            "version >= 2.4 and ModSecurity 2")

    config_path = HostingPanel().get_app_specific_waf_config()
    try:
        hashsum_old = file_hash(config_path, hashlib.sha256)
    except FileNotFoundError:
        hashsum_old = None
    atomic_rewrite(config_path, _rules_config(),
                   backup=hashsum_old is not None)
    hashsum_new = file_hash(config_path, hashlib.sha256)

    def revert():
        try:
            os.rename(config_path + BACKUP_EXTENSION, config_path)
        except FileNotFoundError:
            # truncate file if backup does not exist
            open(config_path, 'w').close()

    def remove_backup():
        with suppress(FileNotFoundError):
            os.unlink(config_path + BACKUP_EXTENSION)

    if hashsum_old != hashsum_new:
        try:
            await configtest()
        except CheckRunError as e:
            logger.error('WAF Rules Set Config is invalid: %s', e)
            revert()
        else:
            try:
                await graceful_restart()
            except CheckRunError:
                logger.error('Web server failed to start with a new config')
                revert()
                await graceful_restart()
            else:
                logger.info('WAF Rules Set Config was successfully updated')
                remove_backup()
    else:
        logger.info('WAF Rules Set Config is already up to date')
        remove_backup()


def _rules_config():
    mapping_path = os.path.join(Index.files_path(MODSEC), MAPPING_FILE)
    with open(mapping_path, encoding='utf-8') as f:
        tags = json.load(f)
    rules_list = map_components_versions_to_tags(
        COMPONENTS_VERSION_DB, tags)
    # sort rules list to get the equal hash for the same set of rules
    rules_config_text = RULES_CONF_PATTERN.format(
        "\n".join(sorted(rules_list)))
    return rules_config_text


async def try_restore_config_from_backup():
    """
        In case if Agent starts and config backup is present — then
    it is required to restore it: original .conf files must be replaced by
    the backup file and then WS restart command must be applied.
    In this case we assume that backed up configuration file is correct
    and none config checks are performed.
    """

    try:
        config_path = HostingPanel().get_app_specific_waf_config()
    except NotImplementedError:
        return

    if os.path.isfile(config_path + BACKUP_EXTENSION):
        os.rename(config_path + BACKUP_EXTENSION, config_path)

        try:
            await graceful_restart()
        except CheckRunError:
            logger.exception(
                'Web server failed to start with a backed up config')
