"""
Adapted script from
raw.githubusercontent.com/iseletsk/kernelchecker/master/py/kc-compat.py
"""
import asyncio
import os
from hashlib import sha1
import logging

import aiohttp

logger = logging.getLogger(__name__)


class KernelUnsupported(Exception):
    pass


def get_kernel_hash():
    with open('/proc/version', 'rb') as f:
        return sha1(f.read()).hexdigest()


def inside_vz_container():
    """
    determines if we are inside Virtuozzo container
    :return: True if inside container, false otherwise
    """
    return (os.path.exists('/proc/vz/veinfo') and
            not os.path.exists('/proc/vz/version'))


def inside_lxc_container():
    return '/lxc/' in open('/proc/1/cgroup').read()


async def is_supported_kernel():
    """Whether the patch server knows about the kernel.

    :raises: aiohttp.ClientError, asyncio.TimeoutError: if patches
    server is unreachable

    """
    url = 'https://patches.kernelcare.com/' + get_kernel_hash() + '/version'
    async with aiohttp.ClientSession() as session:
        async with session.head(url) as resp:
            supported = resp.status < 400
            if not supported:
                logger.info("Response to url '%r' has the status code - %d",
                            url, resp.status)
            return supported


async def check_kc_compat():
    if inside_vz_container() or inside_lxc_container():
        raise KernelUnsupported("Kernel is unsupported: Inside container!")
    else:
        try:
            supported = await is_supported_kernel()
        except (aiohttp.ClientError, asyncio.TimeoutError):
            raise KernelUnsupported("Patches server is unreachable")
        else:
            if not supported:
                raise KernelUnsupported("Kernel is unsupported!")
