#!/opt/alt/python38/bin/python3

"""
This script enables the functionality of WAF Ruleset Configurator.

AppVersionDetector is used to fill in and update the components_versions
database with the information on which CMS is used in each of the passed
document roots (or directories therein).

If the user has app_specific_rules option enabled we also use this
information to disable sets of ModSecurity rules for each directory
using tags.
"""

import argparse
import asyncio
import json
import sys
from base64 import b64encode
from codecs import encode
from subprocess import PIPE, CalledProcessError, run

from defence360agent.contracts.config import FileBasedResourceLimits, \
    ANTIVIRUS_MODE
from defence360agent.malwarelib.config import PHP_PATH
from defence360agent.utils import resource_limits

APP_VERSION_DETECTOR_CMD = [
    PHP_PATH,
    "-d=extension=phar.so",
    "-d=extension=json.so",
    "-d=extension=pdo.so",
    "-d=extension=pdo_sqlite.so",
    "-d=extension=sqlite3.so",
    "-d=extension=intl.so",
    "/opt/app-version-detector/app-version-detector.phar",
    "--sqlite-db-report="
    "/var/lib/cloudlinux-app-version-detector/components_versions.sqlite3",
    "--stdin-dirs",
    "--pathes-in-base64",
    "--scan-depth=2",
    "--send-stats",
]
APP_VERSION_DETECTOR_INTENSITY_KEY = "app_version_detector"
LIST_DOCROOTS_CMD = ["imunify360-agent", "list-docroots", "--json"]
UPDATE_RULES_CMD = ["imunify360-agent", "rules", "update-app-specific-rules"]


async def update():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--update-modsec-rulesets',
        action='store_true',
        help='Also change ModSecurity rulesets',
    )
    args = parser.parse_args(sys.argv[1:])

    doc_roots = json.loads(
        run(
            LIST_DOCROOTS_CMD, stdout=PIPE, stderr=PIPE, check=True
        ).stdout.decode()
    )["items"]

    if not doc_roots:
        return

    stdin = b"\n".join(
        (b",".join([encode(domain, "idna"), b64encode(doc_root.encode())]))
        for doc_root, domain in doc_roots.items()
    )

    cmd = APP_VERSION_DETECTOR_CMD
    p = await resource_limits.create_subprocess(
        cmd,
        intensity_cpu=FileBasedResourceLimits.CPU,
        intensity_io=FileBasedResourceLimits.IO,
        key=APP_VERSION_DETECTOR_INTENSITY_KEY,
        stdout=PIPE,
        stderr=PIPE,
        stdin=PIPE,
    )
    out, err = await p.communicate(input=stdin)
    if p.returncode != 0:
        raise CalledProcessError(
            p.returncode, cmd, output=out, stderr=err
        )

    if args.update_modsec_rulesets and not ANTIVIRUS_MODE:
        run(
            UPDATE_RULES_CMD,
            stdout=PIPE,
            stderr=PIPE,
            check=True,
        )


def main():
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(update())
    except CalledProcessError as e:
        sys.exit(
            f"Command {e.cmd} returned non-zero code {e.returncode},"
            f'\n\t\tStdout: {e.stdout.decode("utf-8", "backslashreplace")},'
            f'\n\t\tStderr: {e.stderr.decode("utf-8", "backslashreplace")}\n'
        )
    finally:  # close loop before exit
        loop.close()


if __name__ == "__main__":
    main()
