import logging
from pathlib import Path

from defence360agent.contracts.config import Malware as Config
from defence360agent.malwarelib.config import MalwareHitStatus
from defence360agent.malwarelib.subsys.malware import MalwareAction
from defence360agent.malwarelib.utils import hash_path, safe_fileops

logger = logging.getLogger(__name__)


class CleanupStorage:
    """
    Store files before cleanup and restore them by request
    """

    path: Path = Path(Config.CLEANUP_STORAGE)

    @staticmethod
    async def _copy(src: Path, dst: Path, safe_src=False, safe_dst=False):
        await safe_fileops.safe_move(
            str(src),
            str(dst),
            src_unlink=False,
            dst_overwrite=True,
            safe_src=safe_src,
            safe_dst=safe_dst,
        )

    @classmethod
    def storage_name(cls, filename: str) -> str:
        """
        Get file name for cleanup storage
        :return: file name
        """
        return hash_path(filename)

    @classmethod
    def get_hit_store_path(cls, hit):
        origin_file = hit.orig_file or hit.true_path
        return cls.path / cls.storage_name(origin_file)

    @classmethod
    async def store(cls, hit):
        dst = cls.get_hit_store_path(hit)
        await cls._copy(hit.true_path, dst, safe_src=False, safe_dst=True)

    @classmethod
    async def store_all(cls, hits):
        if not cls.path.exists():
            cls.path.mkdir(0o700)

        succeeded, not_exist, failed = set(), set(), set()

        for hit in hits:
            try:
                await cls.store(hit)
                succeeded.add(hit)
            except FileNotFoundError:
                not_exist.add(hit)
            except (OSError, safe_fileops.UnsafeFileOperation) as e:
                logger.warning("Failed to store file before cleanup: %r -- %s",
                               str(hit), e)
                failed.add(hit)
                await MalwareAction.cleanup_failed_store(hit.orig_file,
                                                         hit.user)

        return succeeded, failed, not_exist

    @classmethod
    async def restore(cls, hit, safe_dst=False):
        src = cls.path / cls.storage_name(hit.orig_file)
        dst = Path(hit.orig_file)
        await cls._copy(src, dst, safe_src=True, safe_dst=safe_dst)

    @classmethod
    async def restore_all(cls, hits):
        succeeded, failed = set(), set()

        for hit in hits:
            try:
                await cls.restore(hit)
                await MalwareAction.cleanup_restored_original(hit.orig_file,
                                                              hit.user)
                succeeded.add(hit)
            except (OSError, safe_fileops.UnsafeFileOperation) as e:
                await MalwareAction.cleanup_failed_restore(hit.orig_file,
                                                           hit.user)
                logger.warning("Failed to restore file: %r -- %s", str(hit), e)
                failed.add(hit)

        return succeeded, failed

    @classmethod
    async def _clear(cls, path: Path, keep: float) -> bool:
        st = path.stat()
        if st.st_mtime < keep:
            path.unlink()
            return True
        return False

    @classmethod
    async def clear(cls, keep: float) -> int:
        """
        Clear storage
        :param keep: keep files after specified timestamp
        :return:
        """
        cls.path.mkdir(0o700, exist_ok=True)
        cleared = 0
        for path in cls.path.iterdir():
            if await cls._clear(path, keep):
                cleared += 1

        return cleared
