import asyncio
import re
import time
import uuid
from contextlib import suppress
from functools import partial
from logging import getLogger
from typing import Dict, Iterable, List

from defence360agent import utils
from defence360agent.api import inactivity
from defence360agent.contracts.config import Malware as Config
from defence360agent.contracts.hook_events import HookEvent
from defence360agent.contracts.messages import (
    CleanupFailed,
    ConfigUpdate,
    MalwareCleanup,
    MalwareCleanupTask,
    MalwareMRSUpload,
)
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.malwarelib.cleanup.cleaner import (
    CleanupResult,
    MalwareCleaner,
    MalwareCleanupProxy,
)
from defence360agent.malwarelib.cleanup.storage import CleanupStorage
from defence360agent.malwarelib.config import (
    MalwareHitStatus,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHistory, MalwareHit
from defence360agent.malwarelib.subsys.malware import MalwareAction
from defence360agent.utils import recurring_check, split_for_chunk
from defence360agent.utils.common import rate_limit

logger = getLogger(__name__)

DAY = 86400
COUNT_OF_ATTEMPTS_TO_CLEANUP_PER_DAY = 4

_group_by_status = partial(MalwareHit.group_by_attribute, attribute='status')
_group_by_user = partial(MalwareHit.group_by_attribute, attribute='user')

throttled_log_error = rate_limit(
    period=DAY, on_drop=logger.warning)(logger.error)


def filter_cleanable(hits: Iterable[MalwareHit]) -> Iterable:
    return (hit for hit in hits if hit.status == MalwareHitStatus.FOUND)


class Cleanup(MessageSink, MessageSource):
    def __init__(self):
        self._cleanup_task = None
        self._store_original_task = None
        self._running = False
        self._loop = None
        self._sink = None
        self._proxy = None
        self._cleaner = None

    async def create_source(self, loop, sink):
        self._loop = loop
        self._sink = sink
        self._proxy = MalwareCleanupProxy()
        self._cleaner = MalwareCleaner(loop=loop, sink=sink)
        self._cleanup_task = loop.create_task(self.cleanup())

    async def create_sink(self, loop):
        pass

    async def shutdown(self):
        if self._cleanup_task:
            self._cleanup_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._cleanup_task

    @expect(MalwareCleanupTask)
    async def process_cleanup_task(self, message: Dict):
        cause = message.get('cause')
        initiator = message.get('initiator')
        post_action = message.get('post_action')
        scan_id = message.get('scan_id')

        # In case another scan already found some of the hits
        # and the cleanup for them has started.
        hits = MalwareHit.refresh_hits(message['hits'])
        hits = filter_cleanable(hits)
        hits = [
            hit
            for hit in hits
            if hit.resource_type == MalwareScanResourceType.FILE.value
        ]
        hits = await self._filter_failed_to_cleanup_hits(hits)

        self._store_original_task = self._loop.create_task(
            self._store_original(hits, cause, initiator, post_action, scan_id)
        )

    @staticmethod
    async def _filter_failed_to_cleanup_hits(hits: list) -> list:
        """
        Don't try to cleanup the same hit more than
        *COUNT_OF_ATTEMPTS_TO_CLEANUP_PER_DAY*
        """
        hits_to_clean = []
        if hits:
            yesterday = time.time() - DAY
            failed_cleanup_count = {}
            for hits_chunk in split_for_chunk(hits, chunk_size=200):
                failed_cleanup_count.update(
                    dict(
                        MalwareHistory.get_failed_cleanup_events_count(
                            [hit.orig_file for hit in hits_chunk],
                            since=yesterday
                        )
                    )
                )
                await asyncio.sleep(0)
            for hit in hits:
                failures = failed_cleanup_count.get(hit.orig_file, 0)
                if failures >= COUNT_OF_ATTEMPTS_TO_CLEANUP_PER_DAY:
                    throttled_log_error(
                        "Skip cleanup file '%s', since there are too many "
                        "attempts to cleanup it for the last day [%s]",
                        hit.orig_file, failures
                    )
                    continue
                hits_to_clean.append(hit)
        return hits_to_clean

    async def _store_original(
        self, hits, cause, initiator, post_action, scan_id
    ):
        MalwareHit.set_status(hits, MalwareHitStatus.CLEANUP_STARTED)
        original_status = _group_by_status(hits)
        with inactivity.track.task("cleanup_storage"):
            succeeded, failed, not_exist = await CleanupStorage.store_all(hits)
        for hit in failed:
            await self._sink.process_message(
                CleanupFailed(
                    message="Failed to store the original"
                    " from {} to {}".format(
                        hit.orig_file, CleanupStorage.path
                    ),
                    timestamp=int(time.time()),
                )
            )
        self._proxy.add(cause, initiator, post_action, scan_id, succeeded)
        for status, hit_list in original_status.items():
            MalwareHit.set_status([h for h in failed if h in hit_list], status)
        MalwareHit.delete_instances(not_exist)
        await MalwareAction.cleanup_failed(
            not_exist, cause=cause, initiator=initiator
        )

    @staticmethod
    def _user_hits(hits):
        user_hits = _group_by_user(hits)
        return user_hits

    def _cloud_assisted_hits(self):
        action_hits = self._proxy.flush()

        for cause, initiator, post_action, scan_id, all_hits in action_hits:
            blacklist = [
                hit
                for hit in all_hits
                if re.match(r"\w+-BLKH-|cloudhash\.|cld-", hit.type)
            ]
            regular_hits = [hit for hit in all_hits if hit not in blacklist]
            yield (
                regular_hits,
                blacklist,
                cause,
                initiator,
                post_action,
                scan_id,
            )

    async def _start_hook(self, cleanup_id, started, hits):
        dump = [hit.as_dict() for hit in hits]
        cleanup_started = HookEvent.MalwareCleanupStarted(
            cleanup_id=cleanup_id,
            started=started,
            total_files=len(hits),
            DUMP=dump,
        )
        await self._sink.process_message(cleanup_started)

    async def _clean_files(self, hits, blacklist=None, cause=None,
                           initiator=None, post_action=None, scan_id=None):
        user_hits = self._user_hits(hits)
        user_hits_black = self._user_hits(blacklist or [])

        for user in {*user_hits, *user_hits_black}:
            hits_regular = user_hits.get(user, [])
            hits_black = user_hits_black.get(user, [])
            user_hits_all = hits_regular + hits_black

            files = [str(hit.true_path) for hit in hits_regular]
            black = [str(hit.true_path) for hit in hits_black]

            logger.debug('Cleaning files: %s', files + black)
            cleanup_id = uuid.uuid4().hex
            started = time.time()
            await self._start_hook(cleanup_id, started, user_hits_all)
            result, error, cmd = await self._cleaner.start(
                user, files, soft=Config.CLEANUP_TRIM, blacklist=black,
            )
            await self._sink.process_message(
                MalwareCleanup(
                    hits=user_hits_all,
                    result=result,
                    cleanup_id=cleanup_id,
                    started=started,
                    error=error,
                    cause=cause,
                    initiator=initiator,
                    post_action=post_action,
                    scan_id=scan_id,
                    args=cmd,
                )
            )

    async def _cleanup(self):
        if self._running:
            return
        if not self._proxy.hits:
            self._proxy.reset()
            return

        self._running = True

        with inactivity.track.task("cleanup"):
            try:
                data = self._cloud_assisted_hits()
                for (all_hits, blacklist, cause, initiator, post_action,
                     scan_id) in data:
                    await self._clean_files(all_hits, blacklist=blacklist,
                                            cause=cause, initiator=initiator,
                                            post_action=post_action,
                                            scan_id=scan_id)
            finally:
                self._running = False

    @recurring_check(1)
    async def cleanup(self):
        await self._cleanup()


class ResultProcessor(MessageSink, MessageSource):
    async def create_sink(self, loop):
        pass

    async def create_source(self, loop, sink):
        self._sink = sink

    @expect(MalwareCleanup)
    async def store_result(self, message):
        hits: List[MalwareHit] = message["hits"]
        result: CleanupResult = message["result"]
        cause = message.get("cause")
        initiator = message.get("initiator")
        now = time.time()

        processed = [hit for hit in hits if hit in result]
        unprocessed = [hit for hit in hits if hit not in result]
        await MalwareAction.cleanup_unable(unprocessed, cause=cause,
                                           initiator=initiator)

        failed = [hit for hit in processed if result[hit].is_failed()]
        await MalwareAction.cleanup_failed(failed, cause=cause,
                                           initiator=initiator)

        cleaned = [hit for hit in processed if result[hit].is_cleaned()]
        await MalwareAction.cleanup_done(cleaned, cause=cause,
                                         initiator=initiator)
        MalwareHit.set_status(cleaned, MalwareHitStatus.CLEANUP_DONE, now)

        removed = [hit for hit in processed if result[hit].is_removed()]
        await MalwareAction.cleanup_removed(removed, cause=cause,
                                            initiator=initiator)
        MalwareHit.set_status(removed, MalwareHitStatus.CLEANUP_REMOVED, now)

        not_exist = [hit for hit in processed if result[hit].not_exist()]
        MalwareHit.delete_instances(not_exist)

        for status, hit_list in _group_by_status(unprocessed, failed).items():
            MalwareHit.set_status(hit_list, status)

        message['hits'] = MalwareHit.refresh_hits(hits)

        await self.send_failed_to_cleanup_hits_to_mrs(failed)

        return message

    async def send_failed_to_cleanup_hits_to_mrs(
        self, failed_to_cleanup_hits
    ):
        if failed_to_cleanup_hits:
            await self._sink.process_message(
                MalwareMRSUpload(
                    suspicious_hits=[
                        str(hit.true_path) for hit in failed_to_cleanup_hits
                    ],
                    upload_reason="cleanup_failure_current"
                )
            )
            await self._sink.process_message(
                MalwareMRSUpload(
                    suspicious_hits=[
                        str(CleanupStorage.get_hit_store_path(hit))
                        for hit in failed_to_cleanup_hits
                    ],
                    upload_reason="cleanup_failure_original"
                )
            )


class StorageController(MessageSink):
    """Remove old backed up files from storage"""
    def __init__(self):
        self._clear_task = None
        self._keep = Config.CLEANUP_KEEP

    async def create_sink(self, loop):
        self._clear_task = loop.create_task(self.daily_clear())

    async def shutdown(self):
        if self._clear_task:
            self._clear_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._clear_task

    async def _clear(self):
        now = time.time()
        keep_hits = now - self._keep * DAY
        keep_orig = now - (self._keep + 1) * DAY  # keep files one more day
        MalwareHit.delete().where(MalwareHit.cleaned_at < keep_hits).execute()
        cleared = await CleanupStorage.clear(keep_orig)
        if cleared:
            logger.info("Cleanup storage have cleaned. "
                        "Files removed: %s", cleared)

    @expect(ConfigUpdate)
    @utils.log_error_and_ignore()
    async def config_updated(self, _):
        if self._keep != Config.CLEANUP_KEEP:
            self._keep = Config.CLEANUP_KEEP
            await self._clear()

    @recurring_check(DAY)
    async def daily_clear(self):
        await self._clear()
