import functools
import glob
import json
import os
import re
from collections import defaultdict
from enum import Enum
from logging import getLogger
from typing import Any, List, Union

import defence360agent.internals.logger
from defence360agent.contracts.messages import MalwareScan
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.malwarelib.config import (
    MalwareEvent,
    MalwareEventPostponed,
    MalwareHitStatus,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHit, MalwareHitAlternate
from defence360agent.malwarelib.model import MalwareScan as MalwareScanModel
from defence360agent.malwarelib.plugins.detached_scan import (
    MalwareScanMessageInfo,
)
from defence360agent.malwarelib.subsys.malware import MalwareAction
from defence360agent.malwarelib.utils.user_list import panel_users
from defence360agent.model.simplification import run_in_executor
from defence360agent.utils import nice_iterator

logger = getLogger(__name__)


class MalwareScanJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Enum):
            return o.value
        return super().default(o)


class StoreMalwareHits(MessageSink, MessageSource):
    malware_action = MalwareAction
    _loop, _sink = None, None

    async def create_source(self, loop, sink):
        self._loop = loop
        self._sink = sink

    async def create_sink(self, loop):
        pass

    @expect(MalwareScan, async_lock=False)
    async def process_hits(self, message):
        """ MalwareScan is saved to DB when:
         1. Detached scan started - message has no results
         2. Any scan finished - message has summary and results
        Message without summary means that detached scan is finished
         and summary will arrive along with results in another message.
        """
        if not message['summary'].get('path'):
            return

        await self._store_scan(message)

    @expect(MalwareScan)
    async def store_log(self, message):
        with defence360agent.internals.logger.openMalwareScanLog() as logf:
            json.dump(
                dict(summary=message["summary"]),
                logf,
                indent=2,
                sort_keys=False,
                cls=MalwareScanJSONEncoder,
            )

    @staticmethod
    def _store_hit(scanid, filename, status, resource_type, data):
        return MalwareHit.create(
            scanid=scanid,
            resource_type=resource_type,
            user=data['owner'],
            size=data['size'],
            hash=data['hash'],
            orig_file=filename,
            type=data['hits'][0]['matches'],
            timestamp=data['hits'][0]['timestamp'],
            status=status,
            malicious=not data['hits'][0]['suspicious'],
        )

    @staticmethod
    def get_outdated_entries(path_obj: Union[str, list]):
        """
        Return files that may already not be infected, yet we still
        consider them such.

        For example, an infected file might have been removed manually.
        """
        possibly_infected_statuses = (
            MalwareHitStatus.FOUND,
        )
        if isinstance(path_obj, str):
            for path in glob.iglob(path_obj):
                path = os.path.realpath(path)
                if (
                    os.path.isfile(path)
                    and MalwareHit.select()
                    .where(
                        (MalwareHit.orig_file == path)
                        & (MalwareHit.status.in_(possibly_infected_statuses))
                    )
                    .first()
                ):
                    yield path
                else:
                    scanned_dir = re.escape(path) + r"(/.*|\b)"
                    yield from (
                        i.orig_file
                        for i in MalwareHit.select().where(
                            (MalwareHit.orig_file.regexp(scanned_dir))
                            & (
                                MalwareHit.status.in_(
                                    possibly_infected_statuses
                                )
                            )
                        )
                    )
        elif isinstance(path_obj, list):
            yield from iter(path_obj)

    async def _store_scan(self, message):
        """ Process scan message results.

            message: MalwareScan message
        """
        summary = message["summary"]
        if not summary["started"]:
            # Scan is queued/aborted.
            return
        _panel_users = list(await panel_users())

        message_type = MalwareScanMessageInfo(message)
        if message_type.is_summary:
            if not (
                MalwareScanModel.select()
                .where(MalwareScanModel.scanid == message["summary"]["scanid"])
                .exists()
            ):
                scan = MalwareScanModel.create(
                    **summary, resource_type=MalwareScanResourceType.FILE.value
                )
                scan.total_malicious = 0
                scan.save()
            else:
                logger.warning(
                    "Scan %s already in database", message["summary"]["scanid"]
                )
        else:
            results = message["results"]
            await self._store_scan_from_results(results, summary, _panel_users)

    @classmethod
    def _delete_outdated_entries(cls, summary: dict) -> None:
        file_patterns = summary.pop('file_patterns', None)
        exclude_patterns = summary.pop('exclude_patterns', None)
        if (
            summary.get('error') is None
            and file_patterns is None
            and exclude_patterns is None
        ):
            outdated_entries = cls.get_outdated_entries(summary['path'])
            MalwareHit.delete_hits(outdated_entries)

    @staticmethod
    async def _process_default_action_results(
        hit_data, default_action_results
    ):
        pass

    async def _store_scan_from_results(
        self, results: dict, summary: dict, _panel_users: List[dict]
    ):
        scan_id = summary["scanid"]
        scan, created = MalwareScanModel.get_or_create(
            scanid=scan_id,
            defaults={
                **summary,
                "resource_type": MalwareScanResourceType.FILE.value,
            },
        )
        if not created:
            # Detached scan only (second message).
            # Update completed time if scan already exists.
            scan.completed = summary["completed"]

        # get('path') indicates that this is the second message,
        # even if they are out of order
        if results is not None and summary.get("path") is not None:
            self._delete_outdated_entries(summary)

        hits = {
            hit.orig_file: hit
            for hit in MalwareHit.get_hits(files=list(results))
        }
        postponed_hits = defaultdict(list)  # type: dict
        total_malicious = 0

        def _hit_status_race_detected(hit: MalwareHit, detected_timestamp):
            return (
                hit.status == MalwareHitStatus.CLEANUP_STARTED
                or hit.status
                in (
                    MalwareHitStatus.CLEANUP_DONE,
                    MalwareHitStatus.CLEANUP_REMOVED,
                )
                and hit.cleaned_at > detected_timestamp
            )

        # ignore hits are already processed by another scan
        # to avoid send its to CH multiple times
        async for file in nice_iterator(tuple(results.keys())):
            if (
                file in hits
                and _hit_status_race_detected(
                    hits[file], results[file]["hits"][0]["timestamp"]
                )
            ):
                results.pop(file)

        malicious_hits = [
            MalwareHitAlternate.create(scan.scanid, file, data)
            for file, data in results.items()
            if not data["hits"][0]["suspicious"]
        ]

        action_results = await self.malware_action.apply_default_action(
            hits=malicious_hits, cause=summary["type"], sink=self._sink,
        )

        apply_dict = {}
        for hit_info, event, action, try_restore in action_results:
            apply_dict[hit_info.orig_file] = (event, action, try_restore)

        for file, data in results.items():
            status = MalwareHitStatus.FOUND
            result = None
            if file in apply_dict:
                (result, default_action, try_restore,) = apply_dict[file]

                # sent to CH
                data["default_action"] = default_action
                data["try_restore"] = try_restore

                total_malicious += 1

                if isinstance(result, MalwareEvent):
                    if result.malware_eliminated:
                        continue

            hit = await run_in_executor(
                self._loop,
                functools.partial(
                    self._store_hit,
                    scan.scanid,
                    file,
                    status,
                    MalwareScanResourceType.FILE.value,
                    data,
                ),
            )

            if isinstance(result, MalwareEventPostponed):
                key = (
                    result.message,
                    (result.cause, result.initiator, result.post_action),
                )
                postponed_hits[key].append(hit)

        scan.total_malicious = total_malicious
        scan.total_resources = summary["total_files"]
        scan.save()

        if self._sink:
            for (
                (msg_cls, (cause, initiator, post_action)),
                hits,
            ) in postponed_hits.items():
                await self._sink.process_message(
                    msg_cls(
                        hits=hits,
                        scan_id=scan_id,
                        cause=cause,
                        initiator=initiator,
                        post_action=post_action,
                    )
                )

        await self._process_default_action_results(
            results,
            {hit.orig_file: event for hit, event, _, _ in action_results},
        )
