from defence360agent.contracts.messages import MalwareCleanupTask
from defence360agent.feature_management.constants import AV, FULL
from defence360agent.feature_management.lookup import feature
from defence360agent.internals.the_sink import TheSink
from defence360agent.malwarelib.cleanup.cleaner import MalwareCleanupProxy
from defence360agent.malwarelib.cleanup.storage import CleanupStorage
from defence360agent.malwarelib.config import (
    MalwareHitStatus,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHit
from defence360agent.malwarelib.utils.endpoints import MaliciousEndpointStatus
from defence360agent.rpc_tools.lookup import (
    CommonEndpoints,
    RootEndpoints,
    bind,
)
from defence360agent.rpc_tools.validate import validate_av_plus_license
from defence360agent.utils import antivirus_mode


@feature(AV, [FULL])
class CleanupEndpoints(CommonEndpoints):
    @bind('malware', 'malicious', 'cleanup')
    @validate_av_plus_license
    async def cleanup(self, ids, user=None):
        hits = MalwareHit.malicious_select(ids, user=user, cleanup=True)
        if hits:
            await self._sink.process_message(MalwareCleanupTask(hits=hits))

    @bind('malware', 'malicious', 'cleanup-all')
    @validate_av_plus_license
    async def cleanup_all(self, user=None):
        hits = MalwareHit.malicious_select(user=user, cleanup=True)
        if hits:
            await self._sink.process_message(MalwareCleanupTask(hits=hits))

    @bind('malware', 'malicious', 'restore-original')
    @validate_av_plus_license
    async def restore_original(self, ids, user=None):
        hits = MalwareHit.malicious_select(ids, user=user, restore=True)
        succeded, failed = await restore_hits(hits, self._sink)
        return MaliciousEndpointStatus(succeded, failed)


class CleanupRootEndpoints(RootEndpoints):
    def __init__(self, sink):
        super().__init__(sink)
        self._proxy = MalwareCleanupProxy()

    @bind('malware', 'cleanup', 'status')
    async def cleanup_status(self):
        progress = self._proxy.get_progress()
        if progress is None:
            status = {'status': 'stopped'}
        else:
            status = {'status': 'running', 'progress': progress}
        return {'items': status}

    @bind('malware', 'user', 'cleanup')
    @validate_av_plus_license
    async def user_cleanup(self, user):
        hits = MalwareHit.malicious_select(user=user, cleanup=True)
        if hits:
            await self._sink.process_message(MalwareCleanupTask(hits=hits))

    @bind('malware', 'user', 'restore-original')
    @validate_av_plus_license
    async def user_restore_original(self, user):
        hits = MalwareHit.malicious_select(user=user, restore=True)
        succeded, failed = await restore_hits(hits, self._sink)
        return MaliciousEndpointStatus(succeded, failed)


async def restore_hits(hits, sink: TheSink):
    file_hits = [
        hit
        for hit in hits
        if hit.resource_type == MalwareScanResourceType.FILE.value
    ]
    succeeded, failed = await CleanupStorage.restore_all(file_hits)
    MalwareHit.set_status(succeeded, MalwareHitStatus.FOUND)

    if antivirus_mode.disabled:
        import im360.malwarelib.scan.mds.restore

        await im360.malwarelib.scan.mds.restore.restore_hits(hits, sink)

    # FIXME: we cannot include db hits here
    return succeeded, failed
