import base64
import binascii
import csv
import os
from collections import namedtuple
from contextlib import suppress
from pathlib import Path
from time import time

from . import AIBOLIT

SECTIONS = {
    "p": "php_malware",
    "j": "js_malware",
    "s": "suspicious",
    "v": "vulners",
    "c": "cloudhash",
}

SUSPICIOUS_SECTIONS = {"vulners", "suspicious"}

AiBolitCSVReport = namedtuple(
    "AiBolitCSVReport",
    [
        "section",
        "path",
        "signature",
        "ctime",
        "mtime",
        "size",
        "etime",
        "signature_id",
        "hash",
        "signature_name",
        "sha256",
    ],
)


def parse_report_csv(report_path: Path):
    with report_path.open(newline="") as report_stream:
        for raw_row in csv.reader(report_stream, delimiter=","):
            row = AiBolitCSVReport(*raw_row)
            try:
                section = SECTIONS[row.section]
            except KeyError:
                continue
            sig = row.signature_name or "{}.{}".format(
                section, row.signature_id
            )
            timestamp = (
                int(float(row.etime)) if row.section != "v" else int(time())
            )
            file_name = row.path
            with suppress(binascii.Error):
                file_name = base64.b64decode(file_name, validate=True)
            file_name = os.fsdecode(file_name)
            yield {
                "name": AIBOLIT,
                "file_name": file_name,
                "signature": sig,
                "modification_time": int(row.mtime),
                "suspicious": section in SUSPICIOUS_SECTIONS,
                "size": int(row.size or 0),
                "hash": row.sha256 or row.hash or None,
                "timestamp": timestamp,
            }


def parse_report_json(report, base64_path=True):
    for section in SECTIONS.values():
        for hit in report.get(section, []):
            sig = hit.get("sn") or ".".join([section, str(hit["sigid"])])
            # vulners section does not provide timestamp ('et' field)
            #  so current time is used instead.
            #  'et' - time when the file was scanned
            timestamp = (
                int(float(hit["et"]))
                if section != "vulners"
                else int(time())
            )
            file_name = hit["fn"]
            if base64_path:
                with suppress(binascii.Error):
                    file_name = base64.b64decode(file_name, validate=True)
            file_name = os.fsdecode(file_name)
            yield {
                "name": AIBOLIT,
                "file_name": file_name,
                "signature": sig,
                "suspicious": section in SUSPICIOUS_SECTIONS,
                "size": hit["sz"],
                "modification_time": hit["mt"],
                # 'hash' field is still used in 'cloudhash' section
                "hash": hit.get("sha256", hit.get("hash")),
                "timestamp": timestamp,
            }
