from asyncio import gather
from datetime import timedelta

from defence360agent.internals.global_scope import g
from defence360agent.malwarelib.plugins.scan_queue import QueueSupervisorSync
from defence360agent.malwarelib.scan.ai_bolit.detached import (
    AiBolitDetachedScan,
)
from defence360agent.mr_proper import BaseCleaner
from defence360agent.utils import antivirus_mode


class UncompletedScansCleaner(BaseCleaner):
    """
    Cleaner for uncompleted (including aborted) detached scans.
    """

    PERIOD = timedelta(days=1).total_seconds()

    if antivirus_mode.enabled:
        _DETACHED_OPERATIONS = [
            AiBolitDetachedScan,
        ]
    else:
        from im360.malwarelib.scan.mds.detached import (
            MDSDetachedCleanup,
            MDSDetachedRestore,
            MDSDetachedScan,
        )

        _DETACHED_OPERATIONS = [
            AiBolitDetachedScan,
            MDSDetachedCleanup,
            MDSDetachedRestore,
            MDSDetachedScan,
        ]

    @classmethod
    async def cleanup(cls) -> None:
        # We need to bring the queue to a consistent state on agent start.
        # Running 'check_detached_operation' and '_recheck_scan_queue' in async
        # manner leads to already started scan re-running (Wrong 'QueuedScan'
        # state on 'recheck')

        coros = [
            op.check_detached_operation_dir(g.sink)
            for op in cls._DETACHED_OPERATIONS
        ]
        await gather(*coros)

        # We could do a recheck in case some of the scans are 'ABORTED' since
        #  handle_aborted_process is called for them all, so they are removed
        #  from queue and would not be re-run
        QueueSupervisorSync(sink=g.sink).recheck()
