import json
import logging
import os

from urllib.parse import quote_from_bytes, urljoin

from defence360agent import utils
from defence360agent.contracts.config import (
    Core,
    Malware as Config,
)
from defence360agent.contracts.license import LicenseCLN, LicenseError

logger = logging.getLogger(__name__)

_CURL = "/opt/alt/curl/usr/bin/curl"
_ENDPOINT = 'api/v1/upload'
_POST_FILE_TIMEOUT = int(
    os.environ.get("IMUNIFY360_POST_FILE_TIMEOUT", 60 * 60)  # hour
)
FALSE_NEGATIVE = 'false_negative'
FALSE_POSITIVE = 'false_positive'


class ClientError(OSError):
    """HTTP client error.

    It is used to hide what specific http client is used by upload_file().
    """


class FileTooLargeError(OSError):
    pass


def _token_to_headers():
    token = LicenseCLN.get_token()
    headers = {
        'I360-Id': token['id'],
        'I360-Limit': token['limit'],
        'I360-Status': token['status'],
        'I360-Token-Expire-Utc': token['token_expire_utc'],
        'I360-Token-Created-Utc': token['token_created_utc'],
        'I360-Sign': token['sign'],
    }
    for key in headers:
        headers[key] = str(headers[key])

    return headers


async def _post_file(filename, url, headers=None, timeout=None):
    """
    Post *filename* as multipart/form-data to *url* with given HTTP *headers*.

    Return server response as bytes (http body).

    Raise TimeoutError on timeout.
    Raise ConnectionError if failed to connect to host.
    Raise ClientError on error.
    """
    if headers is None:
        headers = {}
    headers_args = [
        b"-H%s: %s" % (header.encode("ascii"), value.encode("latin-1"))
        for header, value in headers.items()
    ]
    quoted_full_path = quote_from_bytes(
        os.fsencode(filename), safe=""
    ).encode("ascii")
    cmd = (
        [os.fsencode(_CURL)]
        + headers_args
        + [b"--max-time", str(timeout).encode("ascii")]
        * (timeout is not None)
        + [
            b"--form",
            # https://curl.haxx.se/docs/knownbugs.html#multipart_formposts_file_name_en
            # https://tools.ietf.org/html/rfc5987
            b'file=@"%s";filename="%s";filename*="UTF-8\'\'%s"'
            % (
                # escape backslash, double quotes
                os.fsencode(filename)
                .replace(b"\\", b"\\\\")
                .replace(b'"', b'\\"'),
                quoted_full_path,
                quoted_full_path,
            ),
            b"--fail",
            # disable progress meter
            b"--silent", b"--show-error",
            url.encode("ascii"),
        ]
    )
    rc, out, err = await utils.run(cmd)
    if rc != 0:
        Error = (
            ConnectionError
            if rc == 7
            else TimeoutError
            if rc == 28
            else ClientError
        )
        raise Error(
            (
                "Failed to post {filename} to {url}:"
                " curl: cmd={cmd}, rc={rc}, out={out}, err={err}"
            ).format(**vars())
        )
    return out


async def upload_file(file: str, upload_reason='unknown'):
    """
    Upload a file to Malware Response Service.
    :param file: path to file
    :param upload_reason: one of 'unknown', 'false_positive', 'false_negative'
    :return: dict representing json response
    :raises LicenseError:
    """
    if not LicenseCLN.is_valid():
        raise LicenseError('File uploading to Malware Responce Serivce '
                           'requires a valid license')
    file_size = os.path.getsize(file)
    if file_size > Config.MAX_MRS_UPLOAD_FILE:
        raise FileTooLargeError(
            'File {} is {} bytes, files larger than {} bytes '
            'are not allowed.'.format(file, file_size,
                                      Config.MAX_MRS_UPLOAD_FILE))

    url = urljoin(Core.API_BASE_URL, _ENDPOINT)
    headers = {
        **_token_to_headers(),
        'I360-Upload-Reason': upload_reason,
    }
    response_body = await _post_file(
        file, url, headers, timeout=_POST_FILE_TIMEOUT
    )
    result = json.loads(response_body.decode())
    logger.info(
        'Uploaded file %r to the Malware Response Service with reason: %s',
        file, upload_reason
    )
    return result
