import csv
import json
import os
import sys
import time
from contextlib import suppress
from pathlib import Path
from typing import Optional

from defence360agent.contracts.config import Core
from defence360agent.utils import snake_case


class RevisiumTempFile:
    ext = "txt"

    def __init__(self, tempdir=None, mode=None):
        self.tempdir = Path(Core.TMPDIR) if tempdir is None else Path(tempdir)
        self.mode = mode

    def __enter__(self):
        filename = "%s_%d.%s" % (self.name, int(time.time() * 10e6), self.ext)
        self._path = self.tempdir / filename
        if self.mode:
            self._path.touch()
            os.chmod(str(self._path), mode=self.mode)
        return self

    def __exit__(self, *_):
        with suppress(FileNotFoundError):
            self._path.unlink()

    @property
    def filename(self):
        return str(self._path)

    @property
    def name(self):
        return snake_case(self.__class__.__name__)


class RevisiumJsonFile(RevisiumTempFile):
    ext = "json"

    def read(self):
        with self._path.open() as f:
            return json.load(f)


class RevisiumCSVFile(RevisiumTempFile):
    ext = "csv"

    @staticmethod
    def _convert_boolean(value: Optional[str]):
        if value:
            low_value = value.lower()
            if low_value in ["true", "false"]:
                return low_value == "true"
        elif value is None:
            return ""
        return value

    def read(self):
        with self._path.open(
            encoding=sys.getfilesystemencoding(),
            errors=sys.getfilesystemencodeerrors(),
        ) as f:
            reader = csv.DictReader(
                f, fieldnames=["f", "e", "d", "s", "t", "r"]
            )
            return [
                {
                    field: self._convert_boolean(value)
                    for field, value in row.items()
                }
                for row in reader
            ]
