import grp
import logging
import os
import pwd
from collections import defaultdict
from typing import NewType, Union

from peewee import fn

from defence360agent.malwarelib.config import (
    MalwareHitStatus,
    MalwareScanResourceType,
    QueuedScanState,
)
from defence360agent.malwarelib.model import MalwareHit, MalwareScan
from defence360agent.subsys.panels import hosting_panel
from defence360agent.utils import get_results_iterable_expression
from defence360agent.utils.threads import to_thread

logger = logging.getLogger(__name__)

Uid = NewType("Uid", int)
Username = NewType("Username", str)


def stub_entry():
    return {
        "user": None,
        "home": None,
        "infected": 0,
        "infected_db": 0,
        "_infected_total": 0,
        "scan_id": None,
        "scan_date": None,
        "scan_status": None,
        "cleanup_status": None,
    }


def system_users():
    """
    Get all system users and initialize a dict for them.

    If a user has leftover config files after being deleted then
    the panel API might treat him as existent. This is resolved by checking
    that a system user is a panel user.
    """
    for entry in pwd.getpwall():
        u = stub_entry()
        u['user'] = entry.pw_name
        u['home'] = entry.pw_dir
        yield u


async def panel_users():
    users = await hosting_panel.HostingPanel().get_users()

    return (u for u in system_users() if u['user'] in users)


def get(user_list, **kwargs) -> dict:
    for u in user_list:
        if all([u[k] == v for k, v in kwargs.items()]):
            return u

    return stub_entry()


def update_infected_count_and_last_scan(user_list):
    homes = [u["home"] for u in user_list]

    def expr(_homes):
        q = (
            MalwareScan.select(
                MalwareScan.scanid, MalwareScan.completed, MalwareScan.path
            )
            .group_by(MalwareScan.path)
            .having(MalwareScan.completed == fn.Max(MalwareScan.completed))
            .where(MalwareScan.path.in_(_homes))
        )
        return q

    # FIXME: refactor this (lots of duplication)
    grouped_hits = (
        MalwareHit.select(MalwareHit.user, fn.COUNT().alias("infected"))
        .where(
            MalwareHit.is_infected()
            & (MalwareHit.resource_type == MalwareScanResourceType.FILE.value)
        )
        .group_by(MalwareHit.user)
    )

    grouped_db_hits = (
        MalwareHit.select(MalwareHit.user, fn.COUNT().alias("infected_db"))
        .where(
            MalwareHit.is_infected()
            & (MalwareHit.resource_type == MalwareScanResourceType.DB.value)
        )
        .group_by(MalwareHit.user)
    )

    grouped_hits_dict = {entry.user: entry.infected for entry in grouped_hits}
    grouped_db_hits_dict = {
        entry.user: entry.infected_db for entry in grouped_db_hits
    }
    actual_scans = get_results_iterable_expression(expr, homes)

    for entry in actual_scans:
        u = get(user_list, home=entry.path)
        u["infected"] = grouped_hits_dict.get(u["user"], 0)
        u["scan_status"] = QueuedScanState.stopped.value
        u["scan_id"] = entry.scanid

    for user, infected_db in grouped_db_hits_dict.items():
        u = get(user_list, user=user)
        u["infected_db"] = infected_db

    for u in user_list:
        u["_infected_total"] = u["infected"] + u["infected_db"]


def update_running_scan_status(user_list, get_scans):
    paths = [u['home'] for u in user_list]
    for scan, status in get_scans(paths):
        u = get(user_list, home=scan.path)
        u['scan_id'] = scan.scanid
        u['scan_status'] = status
        u['scan_type'] = scan.scan_type


def update_cleanup_status(user_list):
    """
    Updates cleanup status for the list of panel users
    If at least on cleanup is running for user then status is 'running'
    Else if there are any finished cleanups then status is 'stopped'
    If no started and finished cleanups then status is not set

    :param user_list:
    """
    users = [u['user'] for u in user_list]

    def expression(users): return MalwareHit.select().where(
        MalwareHit.user.in_(users)).execute()

    for hit in get_results_iterable_expression(expression, users):
        u = get(user_list, user=hit.user)
        if hit.status == MalwareHitStatus.CLEANUP_STARTED:
            u["cleanup_status"] = "running"

        elif u.get("cleanup_status") is None and hit.status in (
            MalwareHitStatus.CLEANUP_DONE,
            MalwareHitStatus.CLEANUP_REMOVED,
        ):
            u["cleanup_status"] = "stopped"


def update_last_scan_date(user_list):
    def expression(homes):
        return (
            MalwareScan.select(MalwareScan.path, MalwareScan.completed)
            .where(MalwareScan.path.in_(homes))
            .group_by(MalwareScan.path)
            .having(MalwareScan.completed == fn.Max(MalwareScan.completed))
        )

    home_to_users = defaultdict(list)
    for user in user_list:
        home_to_users[user["home"]].append(user)
    for scan in get_results_iterable_expression(
        expression, list(home_to_users)
    ):
        for user in home_to_users[scan.path]:
            user["scan_date"] = scan.completed


async def fetch_user_list(get_scans):
    user_list = list(await panel_users())
    update_infected_count_and_last_scan(user_list)
    update_running_scan_status(user_list, get_scans)
    update_cleanup_status(user_list)
    update_last_scan_date(user_list)
    return user_list


def sort(user_list, field="_infected_total", desc=True):
    def getter(element):
        field_type = (
            int
            if field
            in ["infected", "infected_db", "_infected_total", "scan_date"]
            else str
        )
        min_val = chr(0) if field_type == str else 0
        value = element.get(field)
        if value is None:
            value = min_val
        return value

    user_list.sort(key=getter, reverse=desc)
    if field == '_infected_total':
        for user in user_list:
            user.pop("_infected_total")
    return user_list


async def get_file_owner(path, users_from_panel, pw_all):
    """Get username, groupname for file *path*

    pw_all - should contains result of pwd.getpwall()
    user_from_panel - users in current panel (see code comment)

    Returns tuple (user, group, uid, gid) """
    stat = os.stat(path)

    try:
        group = (await to_thread(grp.getgrgid, stat.st_gid)).gr_name
    except KeyError:
        group = stat.st_gid

    # Plesk-panel clients can have two system users with same uids,
    # but only one of them will be used in panel.
    user: Union[Uid, Username] = stat.st_uid
    for pw in pw_all:
        if pw.pw_uid == stat.st_uid and pw.pw_name in users_from_panel:
            user = pw.pw_name
            break

    return user, group, stat.st_uid, stat.st_gid


async def fill_results_owner(results):
    users_from_panel = await hosting_panel.HostingPanel().get_users()
    missing = []
    pw_all = await to_thread(pwd.getpwall)
    for path, data in results.items():
        try:
            user, group, uid, gid = await get_file_owner(
                path,
                users_from_panel,
                pw_all,
            )
        except FileNotFoundError:
            missing.append(path)
        else:
            data["owner"] = user
            data["group"] = group
            data["uid"] = uid
            data["gid"] = gid

    for m in missing:
        del results[m]
