import shutil
import time
import uuid
from pathlib import Path
from typing import Iterable

from defence360agent.contracts.messages import (
    MalwareCleanComplete,
    MalwareCleanup,
    MalwareCleanupTask,
    MalwareRestoreComplete,
)
from defence360agent.contracts.plugins import MessageSink, expect
from defence360agent.internals.global_scope import g
from defence360agent.malwarelib.config import (
    MalwareHitStatus,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHit
from defence360agent.malwarelib.plugins.cleanup import (
    ResultProcessor as ResultProcessorAV,
)
from defence360agent.malwarelib.plugins.cleanup import filter_cleanable, logger
from defence360agent.malwarelib.scan import ScanAlreadyCompleteError
from defence360agent.malwarelib.subsys.malware import MalwareAction
from im360.contracts.messages import (
    MalwareDatabaseCleanup,
    MalwareDatabaseCleanupFailed,
)
from im360.contracts.messages import (
    MalwareDatabaseRestore as MalwareDatabaseRestoreMessage,
)
from im360.contracts.messages import (
    MalwareDatabaseRestoreFailed,
    MalwareDatabaseRestoreTask,
)
from im360.malwarelib.scan.mds.cleaner import MalwareDatabaseCleaner
from im360.malwarelib.scan.mds.detached import (
    MDSDetachedCleanup,
    MDSDetachedRestore,
)
from im360.malwarelib.scan.mds.restore import MalwareDatabaseRestore
from im360.malwarelib.subsys.malware import HackerTrapHitsSaver


class ResultProcessor(ResultProcessorAV):
    """Imunify360 specialization of ResultProcessor, which removes all
    cleaned and removed files from HackerTrap
    """
    @expect(MalwareCleanup)
    async def store_result(self, message):
        message = await super().store_result(message)
        to_remove = [Path(hit) for hit, state in message["result"].items()
                     if (state.is_cleaned() or state.is_removed())]
        await HackerTrapHitsSaver.update_sa_hits([], to_remove)


class CleanupDb(MessageSink):
    def __init__(self):
        self._loop = None

    @staticmethod
    async def _start_cleaner(path, app_name):
        cleanup_id = uuid.uuid4().hex
        await MalwareDatabaseCleaner(cleanup_id, path, app_name).start()

    async def _cleanup_next(self):
        if (
            MalwareHit.db_hits_under_cleanup().exists()
            or (
                next_hit := MalwareHit.db_hits_pending_cleanup()
                .order_by(MalwareHit.timestamp.asc())
                .first()
            )
            is None
        ):
            return
        logger.info(
            "Cleaning hit: (%s::%s)", next_hit.orig_file, next_hit.app_name
        )
        MalwareHit.set_status([next_hit], MalwareHitStatus.CLEANUP_STARTED)
        await self._start_cleaner(next_hit.orig_file, next_hit.app_name)

    async def create_sink(self, loop):
        self._loop = loop
        await self._cleanup_next()

    @expect(MalwareCleanupTask)
    async def process_cleanup_task(self, message):
        hits = MalwareHit.refresh_hits(message["hits"])
        hits_to_clean = filter_cleanable(hits)
        db_hits = [
            hit
            for hit in hits_to_clean
            if hit.resource_type == MalwareScanResourceType.DB.value
        ]
        if not db_hits:
            return

        MalwareHit.set_status(db_hits, MalwareHitStatus.CLEANUP_PENDING)
        await self._cleanup_next()

    @expect(MalwareCleanComplete)
    async def parse_cleanup_results(self, message):
        clean_id = message["scan_id"]
        detached_cleanup = MDSDetachedCleanup(clean_id)
        try:
            cleanup_outcome = await detached_cleanup.complete()
        except ScanAlreadyCompleteError:
            # This happens when AV is woken up by AiBolit. See DEF-11078.
            logger.warning(
                "Cannot complete cleanup %s, assuming it is already complete",
                clean_id,
            )
            return
        finally:
            shutil.rmtree(
                str(detached_cleanup.detached_dir), ignore_errors=True
            )
        await g.sink.process_message(cleanup_outcome)

    @expect(MalwareDatabaseCleanup)
    async def update_cleaned_hits_status(
        self, message: MalwareDatabaseCleanup
    ):
        cleaned_hits = MalwareHit.db_hits_under_cleanup_in(message.succeeded)
        failed_hits = MalwareHit.db_hits_under_cleanup_in(message.failed)
        MalwareHit.set_status(
            cleaned_hits, MalwareHitStatus.CLEANUP_DONE, time.time()
        )
        MalwareHit.set_status(failed_hits, MalwareHitStatus.FOUND)
        await self._cleanup_next()

    @expect(MalwareDatabaseCleanupFailed)
    async def update_failed_hits_status(self, message):
        """
        Clear the queue when the cleanup fails,
        set hits' status back to infected
        """
        # We assume here that all CLEANUP_STARTED hits are part of the
        # same cleanup operation
        hits = MalwareHit.db_hits_under_cleanup()
        MalwareHit.set_status(hits, MalwareHitStatus.FOUND)
        await self._cleanup_next()

    @expect(MalwareDatabaseCleanup)
    async def save_cleanup_events_in_history(
        self, message: MalwareDatabaseCleanup
    ):
        cause = None
        initiator = None
        cleaned_hits = MalwareHit.get_db_hits(message.succeeded)
        await MalwareAction.cleanup_done(
            cleaned_hits, cause=cause, initiator=initiator
        )
        failed_hits = MalwareHit.get_db_hits(message.failed)
        await MalwareAction.cleanup_failed(
            failed_hits, cause=cause, initiator=initiator
        )


class RestoreOriginalDb(MessageSink):
    def __init__(self):
        self.loop = None

    @staticmethod
    async def _restore_next():
        if (
            MalwareHit.db_hits_under_cleanup_restore().exists()
            or (
                hit_to_restore := MalwareHit.db_hits_pending_cleanup_restore()
                .order_by(MalwareHit.timestamp.asc())
                .first()
            )
            is None
        ):
            return
        logger.info(
            "Restoring from cleanup hit: (%s::%s)",
            hit_to_restore.orig_file,
            hit_to_restore.app_name,
        )
        await MalwareDatabaseRestore(
            path=hit_to_restore.orig_file, app_name=hit_to_restore.app_name
        ).restore()
        MalwareHit.set_status(
            [hit_to_restore], MalwareHitStatus.CLEANUP_RESTORE_STARTED
        )

    async def create_sink(self, loop):
        self.loop = loop
        await self._restore_next()

    @staticmethod
    def _filter_under_restore(
        hits: Iterable[MalwareHit],
    ) -> Iterable[MalwareHit]:
        return (
            hit
            for hit in hits
            if hit.status == MalwareHitStatus.CLEANUP_RESTORE_STARTED
        )

    @expect(MalwareDatabaseRestoreTask)
    async def queue_db_restore(self, message):
        MalwareHit.set_status(
            MalwareHit.db_hits()
            .where(MalwareHit.orig_file == message.path)
            .where(MalwareHit.app_name == message.app_name),
            MalwareHitStatus.CLEANUP_RESTORE_PENDING,
        )
        await self._restore_next()

    @expect(MalwareRestoreComplete)
    async def parse_restore_results(self, message):
        restore_id = message["scan_id"]
        detached_restore = MDSDetachedRestore(restore_id)

        try:
            restore_message = await detached_restore.complete()
        except ScanAlreadyCompleteError:
            # This happens when AV is woken up by AiBolit. See DEF-11078.
            logger.warning(
                "Cannot complete restore %s, assuming it is already complete",
                restore_id,
            )
            return
        finally:
            shutil.rmtree(
                str(detached_restore.detached_dir), ignore_errors=True
            )

        await g.sink.process_message(restore_message)

    @expect(MalwareDatabaseRestoreMessage)
    async def update_restored_hits_status(self, message):
        restored_hits = MalwareHit.get_db_hits(message.succeeded)
        MalwareHit.set_status(
            self._filter_under_restore(restored_hits), MalwareHitStatus.FOUND
        )
        await self._restore_next()

    @expect(MalwareDatabaseRestoreMessage)
    async def save_restore_events_in_history(self, message):
        cause = message.get("cause")
        initiator = message.get("initiator")
        restored_hits = MalwareHit.get_db_hits(message.succeeded)
        for hit in restored_hits:
            # FIXME: change cleanup_restored_original to accept multiple
            # values
            await MalwareAction.cleanup_restored_original(
                path=hit.orig_file,
                app_name=hit.app_name,
                resource_type=MalwareScanResourceType.DB.value,
                file_owner=hit.user,
                initiator=initiator,
                cause=cause,
                db_host=hit.db_host,
                db_port=hit.db_port,
                db_name=hit.db_name,
            )
        failed_hits = MalwareHit.get_db_hits(message.failed)
        for hit in failed_hits:
            await MalwareAction.cleanup_failed_restore(
                path=hit.orig_file,
                app_name=hit.app_name,
                resource_type=MalwareScanResourceType.DB.value,
                file_owner=hit.user,
                initiator=initiator,
                cause=cause,
                db_host=hit.db_host,
                db_port=hit.db_port,
                db_name=hit.db_name,
            )

    @expect(MalwareDatabaseRestoreFailed)
    async def update_failed_hits_status(self, message):
        """
        Clear the queue when the restore fails,
        set hits' status back to cleanup_done
        """
        hits = MalwareHit.db_hits_under_restoration()
        MalwareHit.set_status(hits, MalwareHitStatus.CLEANUP_DONE)
        await self._restore_next()
