import pwd
from collections import defaultdict
from logging import getLogger

from defence360agent.contracts.plugins import expect
from defence360agent.malwarelib.config import (
    MalwareEvent,
    MalwareEventPostponed,
    MalwareHitStatus,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHit
from defence360agent.malwarelib.model import MalwareScan as MalwareScanModel
from defence360agent.malwarelib.plugins.store import (
    StoreMalwareHits as StoreMalwareHitsAV,
)
from im360.contracts.messages import MalwareDatabaseScan
from im360.malwarelib.subsys.malware import HackerTrapHitsSaver, MalwareAction

logger = getLogger(__name__)


class StoreMalwareHits(StoreMalwareHitsAV):
    malware_action = MalwareAction

    async def create_sink(self, loop):
        await super().create_sink(loop)
        await HackerTrapHitsSaver.init()

    @staticmethod
    async def _process_default_action_results(
        hit_data, default_action_results
    ):
        """Do additional processing for malicious files"""

        hacker_trap_hits = []
        hacker_trap_sa_hits = []

        for path, data in hit_data.items():
            result = default_action_results.get(path)
            if not isinstance(result, MalwareEvent):
                continue
            if result.malware_eliminated:
                hacker_trap_hits.append(path)

            if any(
                HackerTrapHitsSaver.STANDALONE_MARK in hit["matches"]
                for hit in data["hits"]
            ):
                hacker_trap_sa_hits.append(path)

        await HackerTrapHitsSaver.add_hits(hacker_trap_hits)
        await HackerTrapHitsSaver.update_sa_hits(hacker_trap_sa_hits, [])

    @expect(MalwareDatabaseScan)
    async def store_db_scan(self, message: MalwareDatabaseScan):
        scan = MalwareScanModel.create(
            scanid=message.scan_id,
            started=message.started,
            completed=message.completed,
            type=message.type,
            path=message.path,
            error=message.error,
            total_resources=message.total_resources,
            total_malicious=message.total_malicious,
            resource_type=MalwareScanResourceType.DB.value,
        )
        if not message.hits:
            return
        # FIXME: remove this mapping
        # when we start to store UID instead of username in the db
        uid_to_name = {pw.pw_uid: pw.pw_name for pw in pwd.getpwall()}

        self._delete_outdated_db_entries(message.hits)

        action_results = await self.malware_action.apply_default_action(
            hits=message.hits,
            cause=message.get("type"),
            sink=self._sink,
            resource_type=MalwareScanResourceType.DB.value,
        )

        apply_dict = {}
        for hit, event, action, _ in action_results:
            apply_dict[hit.path] = (event, action)

        postponed_hits = defaultdict(list)  # type: dict

        for hit_info in message.hits:
            result = None
            if hit_info.path in apply_dict:
                (
                    result,
                    default_action,
                ) = apply_dict[hit_info.path]

                # FIXME: DEF-18112 add default_action to hit and send to CH

                if isinstance(result, MalwareEvent):
                    if result.malware_eliminated:
                        continue

            new_hit: MalwareHit = MalwareHit.create(
                scanid=scan,
                user=uid_to_name.get(hit_info.user, hit_info.user),
                orig_file=hit_info.path,
                type=hit_info.signature,
                malicious=True,
                hash=None,
                size=None,
                timestame=None,
                status=MalwareHitStatus.FOUND,
                cleaned_at=None,
                resource_type=MalwareScanResourceType.DB.value,
                app_name=hit_info.app_name,
                db_host=hit_info.db_host,
                db_port=hit_info.db_port,
                db_name=hit_info.db_name,
            )
            if isinstance(result, MalwareEventPostponed):
                key = (
                    result.message,
                    (result.cause, result.initiator, result.post_action),
                )
                postponed_hits[key].append(new_hit)

        if self._sink:
            for (
                    (msg_cls, (cause, initiator, post_action)),
                    hits,
            ) in postponed_hits.items():
                await self._sink.process_message(
                    msg_cls(
                        hits=hits,
                        scan_id=message.scan_id,
                        cause=cause,
                        initiator=initiator,
                        post_action=post_action,
                    )
                )

    @staticmethod
    def _delete_outdated_db_entries(hits):
        orig_files = [hit.path for hit in hits]
        MalwareHit.delete_hits(orig_files)
