import subprocess
from asyncio import CancelledError
from contextlib import suppress
from typing import Iterator

from defence360agent.contracts.config import FileBasedResourceLimits
from defence360agent.malwarelib.config import (
    APP_VERSION_DETECTOR_INTENSITY_KEY,
    PHP_PATH,
)
from defence360agent.malwarelib.scan import ScanFailedError
from defence360agent.utils import resource_limits


class AVDExecutionError(ScanFailedError):
    pass


class AppVersionDetector:
    _cmd = [
        PHP_PATH,
        "-d=extension=phar.so",
        "-d=extension=json.so",
        "-d=extension=pdo.so",
        "-d=extension=pdo_sqlite.so",
        "-d=extension=sqlite3.so",
        "/opt/app-version-detector/app-version-detector.phar",
        # DB is hardcoded in MDS
        "--sqlite-db-report="
        "/var/lib/cloudlinux-app-version-detector/components_versions.sqlite3",
        "--stdin-dirs",
        "--paths-in-base64",
        "--scan-depth=3",
    ]

    def __init__(self):
        self._proc = None

    async def start(self, paths: Iterator[bytes]) -> None:
        stdin = b"\n".join(paths)
        self._proc = await resource_limits.create_subprocess(
            self._cmd,
            intensity_cpu=FileBasedResourceLimits.CPU,
            intensity_io=FileBasedResourceLimits.IO,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            key=APP_VERSION_DETECTOR_INTENSITY_KEY,
        )

        out = err = None
        try:
            out, err = await self._proc.communicate(input=stdin)
        except CancelledError:
            with suppress(ProcessLookupError):
                self._proc.kill()
                await self._proc.wait()

        if self._proc.returncode != 0:
            raise AVDExecutionError(
                message="AVD exited with non-zero return code",
                command=self._cmd,
                return_code=self._proc.returncode,
                out=out,
                err=err,
                path=None,
                scan_id=None,
            )
