import asyncio
import logging
import os
from contextlib import suppress

from defence360agent.contracts.config import MalwareSignatures
from im360.malwarelib.scan.mds import MDS_PATH
from im360.malwarelib.scan.mds.detached import MDSDetachedCleanupDir
from im360.malwarelib.scan.mds.restore import storage_path

logger = logging.getLogger(__name__)


class MalwareDatabaseCleaner:
    def __init__(self, cleanup_id, path, app_name):
        if not path:
            raise ValueError("Cannot instantiate cleaner without path")
        if not app_name:
            raise ValueError("Cannot instantiate cleaner without app names")
        self.cleanup_id = cleanup_id
        self.path = path
        self.app_name = app_name
        self.work_dir = MDSDetachedCleanupDir(self.cleanup_id)

    @property
    def storage_path(self):
        return storage_path(self.path, self.app_name)

    @property
    def cmd(self):
        return [
            "/opt/ai-bolit/wrapper",
            MDS_PATH,
            "--clean",
            "--path",
            self.path,
            "--avdb",
            MalwareSignatures.MDS_AI_BOLIT_HOSTER,
            "--report-file",
            str(self.work_dir.report_file),
            "--detached",
            self.cleanup_id,
            "--progress",
            str(self.work_dir.progress_file),
            "--backup-file",
            self.storage_path,
            "--app-name",
            self.app_name,
            "--procudb",
            MalwareSignatures.MDS_PROCU_DB,
        ]

    async def start(self):
        with self.work_dir:
            with self.work_dir.log_file.open(
                mode="w"
            ) as l_f, self.work_dir.err_file.open(mode="w") as e_f:
                logger.info("Running MDS Cleaner with: %s", self.cmd)

                with suppress(FileNotFoundError):
                    os.unlink(self.storage_path)

                await asyncio.create_subprocess_exec(
                    *self.cmd,
                    start_new_session=True,
                    stdout=l_f,
                    stderr=e_f,
                    cwd=str(self.work_dir),
                )
