import hashlib
import logging
import uuid

from defence360agent.contracts.config import MalwareScanIntensity
from defence360agent.internals.the_sink import TheSink
from defence360agent.malwarelib.config import (
    RESTORE_ORIGINAL_INTENSITY_KEY,
    MalwareScanResourceType,
)
from defence360agent.malwarelib.model import MalwareHit
from defence360agent.malwarelib.utils import hash_path
from defence360agent.utils import resource_limits
from im360.contracts.messages import MalwareDatabaseRestoreTask

from . import MDS_PATH
from .detached import MDSDetachedRestoreDir

logger = logging.getLogger(__name__)


class MalwareDatabaseRestore:
    def __init__(self, *, app_name, path, restore_id=None):
        self.app_name = app_name
        self.path = path
        self.restore_id = restore_id or uuid.uuid4().hex

    def cmd(self, work_dir: MDSDetachedRestoreDir):
        return [
            "/opt/ai-bolit/wrapper",
            MDS_PATH,
            "--restore",
            storage_path(self.path, self.app_name),
            "--path",
            self.path,
            "--app-name",
            self.app_name,
            "--report-file",
            str(work_dir.report_file),
            "--detached",
            self.restore_id,
            "--progress",
            str(work_dir.progress_file),
        ]

    async def restore(self):
        with MDSDetachedRestoreDir(self.restore_id) as work_dir:
            cmd = self.cmd(work_dir)

            with work_dir.log_file.open(
                mode="w"
            ) as l_f, work_dir.err_file.open(mode="w") as e_f:
                logger.info("Running MDS Restore with: %s", cmd)
                await resource_limits.create_subprocess(
                    cmd,
                    intensity_cpu=MalwareScanIntensity.CPU,
                    intensity_io=MalwareScanIntensity.IO,
                    start_new_session=True,
                    stdout=l_f,
                    stderr=e_f,
                    cwd=str(work_dir),
                    key=RESTORE_ORIGINAL_INTENSITY_KEY,
                )


def storage_path(path: str, app_name: str):
    return "/var/imunify360/cleanup_storage/" + hash_path(
        path + app_name, hash_provider=hashlib.sha384
    )


async def restore_hits(hits, sink: TheSink):
    db_hits = [
        hit
        for hit in hits
        if hit.resource_type == MalwareScanResourceType.DB.value
    ]

    hit: MalwareHit
    for hit in db_hits:
        await sink.process_message(
            MalwareDatabaseRestoreTask(
                path=hit.orig_file, app_name=hit.app_name
            )
        )
