import asyncio
import logging
import time
from base64 import b64encode
from glob import iglob
from os import fsencode
from typing import Optional

from defence360agent.contracts.messages import ScanFailed
from defence360agent.internals.global_scope import g
from defence360agent.malwarelib.config import (
    ExitDetachedScanType,
    MalwareScanResourceType,
    MalwareScanType,
)
from defence360agent.malwarelib.scan.queue import QueuedScanBase
from defence360agent.utils import create_task_and_log_exceptions
from im360.malwarelib.scan.app_version_detector import (
    AppVersionDetector,
    AVDExecutionError,
)
from im360.malwarelib.scan.mds.detached import MDSDetachedScan
from im360.malwarelib.scan.mds.scanner import MalwareDatabaseScanner

logger = logging.getLogger(__name__)


class QueuedDbScan(QueuedScanBase):
    resource_type = MalwareScanResourceType.DB

    def __init__(
        self,
        path: str,
        scanid: Optional[str] = None,
        # FIXME: Use Enum instead of a class with str attributes.
        scan_type: str = MalwareScanType.ON_DEMAND,
        created: int = None,
        started: Optional[float] = None,
        intensity: Optional[str] = None,
        home_dirs=None,
        intensity_cpu=None,
        intensity_io=None,
        intensity_ram=None,
        **_,
    ):
        super().__init__(
            path=path,
            scanid=scanid,
            scan_type=scan_type,
            created=created,
            started=started,
            intensity=intensity,
            home_dirs=home_dirs,
            intensity_cpu=intensity_cpu,
            intensity_io=intensity_io,
            intensity_ram=intensity_ram,
        )
        self.detached_scan = MDSDetachedScan(self.scanid)

    def __setstate__(self, state):
        super().__setstate__(state)
        self.detached_scan = MDSDetachedScan(self.scanid)

    @property
    def total_resources(self) -> int:
        return self.detached_scan.total_resources

    async def _scan(self):
        # app-version-detector should recursive check all directories,
        # no need to extract them explicitly.
        # Used to make files and db scans idempotent (DEF-19264)
        # MDS scanner (php) should alerady handle /path/* as /path/ (DEF-19096)
        apps_path = (
            self.path.rstrip('*') if self.path.endswith("/*") else self.path
        )
        unglobbed_paths = [b64encode(fsencode(d)) for d in iglob(apps_path)]
        try:
            await AppVersionDetector().start(unglobbed_paths)
        except AVDExecutionError as exc:  # Exited with non-zero return code
            await g.sink.process_message(
                self._generate_scan_failed_message(
                    exc.command, exc.out, exc.err
                )
            )
        else:
            await MalwareDatabaseScanner(
                self.path,
                [self.path],  # FIXME: pass unglobbed_paths here
                **self.args,
                scan_type=self.scan_type,
                scan_id=self.scanid,
            ).scan()

    async def start(self) -> None:
        self.started = time.time()
        self.scanner_task = create_task_and_log_exceptions(
            asyncio.get_event_loop(), self._scan
        )

    @staticmethod
    def _generate_scan_failed_message(cmd, stdout, stderr):
        """Generate ScanFailed message"""
        msg = ScanFailed(
            out=stdout,
            err=stderr,
            command=" ".join(cmd),
            message=ExitDetachedScanType.ABORTED,
        )
        logger.warning("Scan was aborted: %s, %s", msg["err"], msg["out"])
        return msg


def db_scans(paths, home_dirs, scan_args):
    return [
        QueuedDbScan(path, home_dirs=home_dirs, **scan_args) for path in paths
    ]
